
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysManager.h
**
** DESCRIPTION: This is header file to define the class:
**              CClientFileSysManager
**
**              This class mainly manage all file system information used
**              by one client.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
**  CLASS LIST:
**      CClientFileSysManager
**
**
****************************************************************************/

#ifndef _CLIENTFILESYSMANAGER_H_
#define _CLIENTFILESYSMANAGER_H_

#include "Common.h"
#include "ClientFileSysBase.h"
#include "ClientFileSysInfo.h"

class CClientLoadMgrThread;

#define FILESYS_MAXNUM                  8

#define IOSIZEDIST_MAXITEMNUM           16

typedef struct
{
    int pcnt;       /* percentile */
    int bufs;       /* Block_size KB xfers */
    int frags;      /* boolean - is a frag present */
} SIOSizeDist[IOSIZEDIST_MAXITEMNUM];


class CClientFileSysManager
{
private:
    CClientLoadMgrThread *m_LoadMgrThread;                  /* the load manager thread I am serving for */

    int m_num;                                              /* number of file system */
    CClientFileSysBase *m_fsbase[FILESYS_MAXNUM];           /* array for all base file system information */
    CClientFileSysInfo *m_fsinfo[FILESYS_MAXNUM];           /* array for all file system information */
    int m_range[FILESYS_MAXNUM];                            /* probabilities to select each file system */

    static int m_BlockSize;                                 /* read and write block size in bytes, be multiple of 1024 */
    static SIOSizeDist m_ReadDist;                          /* distribution of size for read operation */
    static SIOSizeDist m_WriteDist;                         /* distribution of size for write operation */
    static int m_MaxBufNum;                                 /* max number of Block_size xfers */
    static double m_AvgOpsPerReadReq;                       /* average number of operation for one read request */
    static double m_AvgOpsPerWriteReq;                      /* average number of operation for one write request */

/* constructor and destructor */
public:
    CClientFileSysManager(CClientLoadMgrThread *pLoadMgrThread);
	~CClientFileSysManager();

/* operations */
public:
    int GetFileSysNum();

    void AddFileSysBase(CClientFileSysBase *fsbase, int prob);
    CClientFileSysBase *GetFileSysBase(int fs);
    void DeleteAllFileSysBase();

    void SetFileSysInfo(int fs, CClientFileSysInfo *fsinfo);
    CClientFileSysInfo *GetFileSysInfo(int fs);
    void DeleteAllFileSysInfo();

    int GetBlockSize();
    double GetAvgOpsPerReadReq();
    double GetAvgOpsPerWriteReq();
    int GetIOOpSize(int op);            /* op=1 for read, op=2 for write */

    int SelectFileSys();

    BOOL GetIOFileInfo(int req_size, char *name, int &fs, int &handle, int &size);
    void SetIOFileInfo(int fs, int handle, int size);

    BOOL GetNonIOFileInfo(int option, char *name, int &fs, int &handle, int &mode, int ref_fs=-1);  /* option: 0(access) 1(remove) 2(create) */
    void SetNonIOFileInfo(int option, int fs, int handle, int mode);                                /* option: 0(access) 1(remove) 2(create) */

    BOOL GetDirInfo(int option, char *name, int &fs, int &handle);                                  /* option: 0(access) 1(remove) 2(create) */
    void SetDirInfo(int option, int fs, int handle);                                                /* option: 0(access) 1(remove) 2(create) */

    BOOL GetSymlinkInfo(int option, char *name, int &fs, int &handle);                              /* option: 0(access) 1(remove) 2(create) */
    void SetSymlinkInfo(int option, int fs, int handle);                                            /* option: 0(access) 1(remove) 2(create) */
};

#endif

