
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysRefThread.h
**
** DESCRIPTION: This is header file to define the class:
**              CClientFileSysRefThread
**
**              This class mainly manage the thread to refresh file system.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
**  CLASS LIST:
**      CClientFileSysRefThread
**
**
****************************************************************************/

#ifndef _CLIENTFILESYSREFTHREAD_H_
#define _CLIENTFILESYSREFTHREAD_H_

#include "Common.h"
#include "BaseThread.h"

class CClientFileSysRefPool;

class CClientFileSysBase;

class CClientFileSysRefThread : public CBaseThread
{
public:
    int m_id;

private:
    CClientFileSysRefPool *m_pool;
    CClientFileSysBase *m_fsbase;

#ifndef WIN32
    pthread_mutex_t m_lock;
    pthread_cond_t m_event;
    BOOL m_ready;
#else
    HANDLE m_readySem;
#endif

    int m_subdir;

    /* the following pointers point to data structure of m_fsbase */
    int *m_IONum;
    int **m_IOSize;
    int *m_FileNum;
    ushort **m_FileMode;
    int *m_DirNum;
    uchar **m_DirMode;
    int *m_SymlinkNum;
    uchar **m_SymlinkMode;

    BOOL *m_foundIO;
    BOOL *m_foundFile;
    BOOL *m_foundDir;
    BOOL *m_foundSymlink;

/* constructor and destructor */
public:
    CClientFileSysRefThread();
    CClientFileSysRefThread(int id, CClientFileSysRefPool *pool, CClientFileSysBase *fsbase);
	~CClientFileSysRefThread();

/* operations */
public:
    void run();

    void SetFSBaseData(int *IONum, int **IOSize, int *FileNum, ushort **FileMode, int *DirNum, uchar **DirMode, int *SymlinkNum, uchar **SymlinkMode);
    void SetFoundNum(int foundIO_Num, int foundFile_Num, int foundDir_Num, int foundSymlink_Num);

    void SetIdle();
    void SetSubdirAndStart(int subdir);
    void Quit();
    void RefSubdir();
};

#endif

