
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientLoadGenThread.cpp
**
** DESCRIPTION: This class mainly manage the thread to generate load from
**              client to file server.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientLoadGenThread::CClientLoadGenThread()
**     CClientLoadGenThread::~CClientLoadGenThread()
**     void CClientLoadGenThread::run()
**
**
****************************************************************************/


#ifndef WIN32
#include <unistd.h>
#endif
#include "Common.h"
#include "ClientLoadGenThread.h"
#include "ClientLoadGenPool.h"
#include "ClientLoadMgrThread.h"

CClientLoadGenThread::CClientLoadGenThread()
{
    m_pool = NULL;
#ifndef WIN32
    pthread_mutex_init(&m_lock, NULL);
    pthread_cond_init(&m_event, NULL);
    m_ready = FALSE;
#else
    m_readySem = CreateSemaphore(NULL, 0, 1, NULL);
#endif
}

CClientLoadGenThread::CClientLoadGenThread(int id, CClientLoadGenPool *pool)
{
    m_id = id;
    m_pool = pool;
#ifndef WIN32
    pthread_mutex_init(&m_lock, NULL);
    pthread_cond_init(&m_event, NULL);
    m_ready = FALSE;
#else
    m_readySem = CreateSemaphore(NULL, 0, 1, NULL);
#endif
}

CClientLoadGenThread::~CClientLoadGenThread()
{
#ifdef WIN32
    if (m_readySem != NULL)
        CloseHandle(m_readySem);
#endif
}

void CClientLoadGenThread::run()
{
    while ( 1 )
    {
        /* wait for ready signal */
#ifndef WIN32
        LockMutex(m_lock);
        while (! m_ready)
            pthread_cond_wait(&m_event, &m_lock);
        m_ready = FALSE;
        UnlockMutex(m_lock);
#else
        WaitForSingleObject(m_readySem, INFINITE);
#endif

        /* do actual work */
        DoOp();

        /* resume to idle state */
        SetIdle();
    }
}

void CClientLoadGenThread::SetIdle()
{
    m_pool->SetIdleLoadGenThread(m_id);
}

void CClientLoadGenThread::SetOpHostAndStart(CClientLoadMgrThread *host)
{
#ifndef WIN32
    LockMutex(m_lock);

    /* set parameter of the operation */
    m_host = host;

    m_ready = TRUE;
    pthread_cond_signal(&m_event);
    UnlockMutex(m_lock);
#else
    m_host = host;
    ReleaseSemaphore(m_readySem, 1, NULL);
#endif
}

void CClientLoadGenThread::DoOp()
{
    char opname[MAX_NAMELEN];
    CClientOperation *op = m_host->GetOperationManager()->GetOp(m_host, opname);

//    printf("Thread %d: do op (%s)\n", m_id, opname);

CSysTime a;
    op->DoOP();
CSysTime b;
b -= a;
if (b.m_sec > 1)
printf("Thread %d: op (%s), %d sec %d usec\n", m_id, opname, b.m_sec, b.m_usec);

    delete op;
}
