
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientLoadMgrThread.h
**
** DESCRIPTION: This is header file to define the class:
**              CClientLoadMgrThread
**
**              This class generates operation and picks a idle loadgen
**              thread to execute the operation.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
**  CLASS LIST:
**      CClientLoadMgrThread
**
**
****************************************************************************/

#ifndef _CLIENTLOADMGRTHREAD_H_
#define _CLIENTLOADMGRTHREAD_H_

#include "Common.h"
#include "BaseThread.h"
#include "ClientLoadGenPool.h"
#include "ClientFileSysManager.h"
#include "ClientOperationManager.h"
#include "SocketClient.h"
#include "NetXDT.h"

//class CClientOperationManager;

class CClientLoadMgrThread : public CBaseThread
{
private:
    CSocketClient *m_socket;

    CClientLoadGenPool *m_LoadGenPool;
    CClientFileSysManager *m_fsinfoMgr;
    CClientOperationManager *m_opMgr;

    char m_myname[MAX_NAMELEN];             /* host name of this client */
    char m_myip[MAX_IPADDRLEN];             /* ip address of this client */

    int m_load_id;                          /* identity of this load */
    int m_client_id;                        /* identity of this client */
    char m_prim[MAX_NAMELEN];               /* prime client name */
    char m_prim_ip[MAX_IPADDRLEN];          /* prime client ip address */
    int m_load;                             /* operation number per second */
    BOOL m_stress_server;                   /* whether stress server (no thinking time) */

    char m_ServerName[MAX_NAMELEN];         /* file server name or ip address */
    char m_ServerPath[MAX_PATHLEN];         /* file system path on file server */
    char m_LocalPath[MAX_PATHLEN];          /* client path which file system is mounted on */
    BOOL m_mount_noac;                      /* mount file system with "noac" option */

    BOOL m_auto_refreshed;                  /* whether to be auto refreshed, check when */
                                            /* receiving new shuffled file system information */

    int m_ops;

#ifndef WIN32
    pthread_mutex_t m_ops_lock;

    pthread_mutex_t m_ready_lock;
    pthread_cond_t m_ready_event;
    BOOL m_ready;

    pthread_mutex_t m_break_lock;
    int m_break;
#else
    HANDLE m_ops_lock;

    HANDLE m_ready_Sem;

    HANDLE m_break_lock;
    int m_break;
#endif
	
/* constructor and destructor */
public:
    CClientLoadMgrThread(CSocketClient *socket);
	~CClientLoadMgrThread();

/* operations */
public:
    CClientFileSysManager *GetFileSysManager();
    CClientOperationManager *GetOperationManager();

    int GetLoadID();
    int GetClientID();
    char *GetPrimName();
    int GetLoad();
    void InitLoadGenThreadPool(int num);

    void SetAutoRefreshed(BOOL auto_ref=TRUE);

    void SetOps(int ops);
    int GetOps();
    void AddOps(int ops);

    void ResetReady();
    BOOL WaitForReady(int timeout); /* return TRUE: read, FALSE: timeout */
    void SetReady();

    void ResetBreak();
    int IsBreak();      // 0: no break   1: break for shuffle   2: exit
    void SetBreak(int reason);

    void AddFileSysBase(CNet_FsSysBase *net_fsbase);
    void UpdateFileSysInfo(int index, CNet_FsSysInfo *net_fsinfo);

    BOOL RecvStartLoadFromPrim(CSocketClient *msg_sock);
    BOOL RecvFSBaseInfoFromPrim(CSocketClient *msg_sock);
    BOOL RecvGoAheadFromPrim(CSocketClient *msg_sock, int timeout); // timeout is in millisecond
    BOOL RecvFileSysInfoFromPrim(CSocketClient *msg_sock);
    BOOL RecvBreakMessageFromPrim(CSocketClient *msg_sock);

    BOOL SendAckToPrim(int ack_code);
    BOOL SendFileSysInfoToPrim();

    int DoOp();         // 1: break for shuffle    2: exit

    void run();
};

#endif

