
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientPrimMgrThread.cpp
**
** DESCRIPTION: This class mainly generate a thread to do the primary client
**              work.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientPrimMgrThread::ClientPrimMgrThread()
**     CClientPrimMgrThread::~ClientPrimMgrThread()
**     void CClientPrimMgrThread::run()
**
**
****************************************************************************/


#ifndef WIN32
#include <unistd.h>
#include <ctype.h>
#endif
#include "Common.h"
#include "ClientPrimMgrThread.h"
#include "SocketClient.h"
#include "Packet.h"
#include "NetXDT.h"
#include "SysTime.h"
#include "Pumper.h"

CClientPrimMgrThread::CClientPrimMgrThread(CSocketClient *socket)
{
    m_socket = socket;

#ifndef WIN32
    pthread_mutex_init(&m_ra_lock, NULL);
    pthread_cond_init(&m_ra_event, NULL);
    m_ra_ready = FALSE;

    InitMutex(m_lc_lock);
    pthread_cond_init(&m_lc_event, NULL);
    m_lc_ready = 0;

    InitMutex(m_ref_lock);
    pthread_cond_init(&m_ref_event, NULL);
    m_ref_ready = 0;
    m_ref_signal = 0;

    InitMutex(m_signal_lock);
    pthread_cond_init(&m_signal_event, NULL);
    m_signal_value = 0;
#else
    m_ra_ready_Sem = CreateSemaphore(NULL, 0, 1, NULL);

    m_lc_ready_Sem = CreateSemaphore(NULL, 0, 1, NULL);
    InitMutex(m_lc_lock);
    m_lc_ready = 0;

    m_ref_ready_Sem = CreateSemaphore(NULL, 0, 1, NULL);
    InitMutex(m_ref_lock);
    m_ref_ready = 0;
    m_ref_signal = 0;

    m_signal_Sem = CreateSemaphore(NULL, 0, 1, NULL);
    InitMutex(m_signal_lock);
    m_signal_value = 0;
#endif

    m_fsnum = 0;

    m_shuffle_time = SHUFFLE_TIME;      /* shuffle every 30 minutes */

//DEBUG
//m_shuffle_time = 40;
}

CClientPrimMgrThread::~CClientPrimMgrThread()
{
    extern CPumper thePumper;

    if (m_socket != NULL)
        delete m_socket;

#ifdef WIN32
    if (m_ra_ready_Sem != NULL)
        CloseHandle(m_ra_ready_Sem);
    if (m_lc_ready_Sem != NULL)
        CloseHandle(m_lc_ready_Sem);
    if (m_lc_lock != NULL)
        CloseHandle(m_lc_lock);
    if (m_ref_ready_Sem != NULL)
        CloseHandle(m_ref_ready_Sem);
    if (m_ref_lock != NULL)
        CloseHandle(m_ref_lock);
    if (m_signal_Sem != NULL)
        CloseHandle(m_signal_Sem);
    if (m_signal_lock != NULL)
        CloseHandle(m_signal_lock);
#endif

    int i, j;
    for (i=0; i<m_fsnum; i++)
    {
        if (m_fs[i] != NULL)
        {
            for (j=0; j<m_fs[i]->SubDirNum; j++)
                delete [] m_fs[i]->IOSize[j];
            for (j=0; j<m_fs[i]->SubDirNum; j++)
                delete [] m_fs[i]->FileMode[j];
            for (j=0; j<m_fs[i]->SubDirNum; j++)
                delete [] m_fs[i]->DirMode[j];
            for (j=0; j<m_fs[i]->SubDirNum; j++)
                delete [] m_fs[i]->SymlinkMode[j];
            delete [] m_fs[i]->IONum;
            delete [] m_fs[i]->IOSize;
            delete [] m_fs[i]->FileNum;
            delete [] m_fs[i]->FileMode;
            delete [] m_fs[i]->DirNum;
            delete [] m_fs[i]->DirMode;
            delete [] m_fs[i]->SymlinkNum;
            delete [] m_fs[i]->SymlinkMode;
            delete [] m_fs[i]->OptDir;

            delete m_fs[i];
        }
    }

    thePumper.DeleteThread(2, this);
}

int CClientPrimMgrThread::GetLoadID()
{
    return m_load_id;
}

int CClientPrimMgrThread::GetStatus()
{
    return m_status;
}

void CClientPrimMgrThread::ResetRaReady()
{
#ifndef WIN32
    LockMutex(m_ra_lock);
    m_ra_ready = FALSE;
    UnlockMutex(m_ra_lock);
#else
#endif
}

void CClientPrimMgrThread::WaitForRaReady()
{
#ifndef WIN32
    LockMutex(m_ra_lock);
    while (! m_ra_ready)
        pthread_cond_wait(&m_ra_event, &m_ra_lock);
    m_ra_ready = FALSE;
    UnlockMutex(m_ra_lock);
#else
    WaitForSingleObject(m_ra_ready_Sem, INFINITE);
#endif
}

void CClientPrimMgrThread::SetRaReady()
{
#ifndef WIN32
    LockMutex(m_ra_lock);
    m_ra_ready = TRUE;
    pthread_cond_signal(&m_ra_event);
    UnlockMutex(m_ra_lock);
#else
    ReleaseSemaphore(m_ra_ready_Sem, 1, NULL);
#endif
}

void CClientPrimMgrThread::ResetLcReady()
{
    LockMutex(m_lc_lock);
    m_lc_ready = 0;
    for (int i=0; i<m_client_num; i++)
        m_lc_status[i] = -1;
    UnlockMutex(m_lc_lock);
}

/* timeout is in milliseconds */
/* return value: TRUE for all lc is ready, FALSE for timeout */
BOOL CClientPrimMgrThread::WaitForLcReady(int *lc_status, int timeout)
{
    BOOL result;

#ifndef WIN32
    struct timespec to;
    to.tv_sec = time(NULL) + timeout / 1000;
    to.tv_nsec = (timeout % 1000) * 1000;
    LockMutex(m_lc_lock);
    while (m_lc_ready < m_client_num)
    {
        if (pthread_cond_timedwait(&m_lc_event, &m_lc_lock, &to) != 0)
            break;  // timeout
    }
    result = (m_lc_ready >= m_client_num);

#else
    if (WaitForSingleObject(m_lc_ready_Sem, timeout) == WAIT_TIMEOUT)
        result = FALSE;
    LockMutex(m_lc_lock);
#endif

    /* reset the status to waiting state */
    for (int i=0; i<m_client_num; i++)
    {
        lc_status[i] = m_lc_status[i];
        m_lc_status[i] = -1;
    }
    m_lc_ready = 0;
    UnlockMutex(m_lc_lock);

    return result;
}

void CClientPrimMgrThread::AddLcReady(int lc_id, int lc_status)
{
    LockMutex(m_lc_lock);
    m_lc_status[lc_id] = lc_status;
    m_lc_ready ++;
#ifndef WIN32
    if (m_lc_ready >= m_client_num)
        pthread_cond_signal(&m_lc_event);
#else
    if (m_lc_ready >= m_client_num)
        ReleaseSemaphore(m_lc_ready_Sem, 1, NULL);
#endif
    UnlockMutex(m_lc_lock);
}

void CClientPrimMgrThread::SetRefClientNum(int ref_client_num)
{
    m_ref_client_num = ref_client_num;
}

void CClientPrimMgrThread::ResetRefReady()
{
    LockMutex(m_ref_lock);
    m_ref_ready = 0;
    m_ref_signal = 0;
    UnlockMutex(m_ref_lock);
}

/* return value: 0:timeout 1:ready 2:client-error 3,4:pause,stop-command */
int CClientPrimMgrThread::TimeWaitForRefReadySignal(int millisecond)
{
#ifndef WIN32
    struct timespec to;
    to.tv_sec = time(NULL) + millisecond / 1000;
    to.tv_nsec = (millisecond % 1000) * 1000;
    LockMutex(m_ref_lock);
    int return_value;
    BOOL IsTimeout = FALSE;
    while (m_ref_ready < m_ref_client_num && m_ref_signal == 0)
    {
        if (pthread_cond_timedwait(&m_ref_event, &m_ref_lock, &to) != 0)
        {
            IsTimeout = TRUE;
            break;  // timeout
        }
    }
    if (IsTimeout)
        return_value = 0;
    else
    {
        if (m_ref_ready >= m_ref_client_num)
        {
            return_value = 1;
            m_ref_ready = 0;
        }
        else
        {
            return_value = m_ref_signal;
            m_ref_signal = 0;
        }
    }
    UnlockMutex(m_ref_lock);

    return return_value;
#else
    int return_value;

    if (WaitForSingleObject(m_ref_ready_Sem, millisecond) == WAIT_TIMEOUT)
        return_value = 0;
    else
    {
        if (m_ref_ready >= m_ref_client_num)
        {
            return_value = 1;
            m_ref_ready = 0;
        }
        else
        {
            return_value = m_ref_signal;
            m_ref_signal = 0;
        }
    }

    return return_value;
#endif
}

void CClientPrimMgrThread::AddRefReady()
{
    LockMutex(m_ref_lock);
    m_ref_ready ++;
#ifndef WIN32
    if (m_ref_ready >= m_ref_client_num)
        pthread_cond_signal(&m_ref_event);
#else
    if (m_ref_ready >= m_ref_client_num)
        ReleaseSemaphore(m_ref_ready_Sem, 1, NULL);
#endif
    UnlockMutex(m_ref_lock);
}

/* sig_code=2:client-error 3,4:pause,stop-command */
void CClientPrimMgrThread::SetRefSignal(int sig_code)
{
#ifndef WIN32
    LockMutex(m_ref_lock);
    m_ref_signal = sig_code;
    pthread_cond_signal(&m_ref_event);
    UnlockMutex(m_ref_lock);
#else
    LockMutex(m_ref_lock);
    m_ref_signal = sig_code;
    UnlockMutex(m_ref_lock);
    ReleaseSemaphore(m_ref_ready_Sem, 1, NULL);
#endif
}

void CClientPrimMgrThread::ResetSignal()
{
    LockMutex(m_signal_lock);
    m_signal_value = 0;
    UnlockMutex(m_signal_lock);
}

int CClientPrimMgrThread::TimeWaitForSignal(int millisecond)
{
#ifndef WIN32
    int sig_code;
    struct timespec to;
    to.tv_sec = time(NULL) + millisecond / 1000;
    to.tv_nsec = (millisecond % 1000) * 1000;
    LockMutex(m_signal_lock);
    while (m_signal_value == 0)
    {
        if (pthread_cond_timedwait(&m_signal_event, &m_signal_lock, &to) != 0)
        {
            m_signal_value = 0;
            break;  // timeout
        }
    }
    sig_code = m_signal_value;
    UnlockMutex(m_signal_lock);
    return sig_code;
#else
    if (WaitForSingleObject(m_signal_Sem, millisecond) == WAIT_TIMEOUT)
        m_signal_value = 0;
    return m_signal_value;
#endif
}

void CClientPrimMgrThread::SetSignal(int sig_code)
{
#ifndef WIN32
    LockMutex(m_signal_lock);
    m_signal_value = sig_code;
    pthread_cond_signal(&m_signal_event);
    UnlockMutex(m_signal_lock);
#else
    LockMutex(m_signal_lock);
    m_signal_value = sig_code;
    UnlockMutex(m_signal_lock);
    ReleaseSemaphore(m_signal_Sem, 1, NULL);
#endif
}

void CClientPrimMgrThread::UpdateLoadInfo(CNet_StartLoadOnPrim *net_startload)
{
    m_load_id = net_startload->id;
    m_client_num = net_startload->num;
    for (int i=0; i<m_client_num; i++)
    {
        strcpy(m_clients[i], net_startload->clients[i]);
        strcpy(m_client_ips[i], net_startload->ipaddrs[i]);
        m_load[i] = net_startload->opnum[i];
    }
    m_stress_server = net_startload->stress_server;
    m_run_time = net_startload->runtime;
    m_used_time = net_startload->usedtime;
    strcpy(m_rmname, net_startload->RMName);
//    strcpy(m_rmip, net_startload->RMIp);
    strcpy(m_rmip, m_socket->GetFromIp());
    if (! isdigit(m_rmip[0]))
        strcpy(m_rmip, net_startload->RMIp);
    strcpy(m_myname, net_startload->PrimName);
    strcpy(m_myip, net_startload->PrimIp);
}

void CClientPrimMgrThread::UpdateFileSysInfo(int client_id, int fs_index, CNet_FsSysInfo *net_fsinfo)
{
    SPrimFileSysInfo *fsinfo = m_fs[fs_index];

    fsinfo->TotalFileNum += net_fsinfo->TotalFileNum;
    fsinfo->TotalDirNum += net_fsinfo->TotalDirNum;
    fsinfo->TotalSymlinkNum += net_fsinfo->TotalSymlinkNum;

    int index, oldnum, dirmodenum, symlinkmodenum;
    int optdirnum = net_fsinfo->OptDirNum;
    for (int i=0; i<optdirnum; i++)
    {
        index = net_fsinfo->OptDir[i];

        oldnum = fsinfo->IONum[index];
        fsinfo->IONum[index] = net_fsinfo->IONum[i];
        if (oldnum != fsinfo->IONum[index])
        {
            delete [] fsinfo->IOSize[index];
            fsinfo->IOSize[index] = new int [fsinfo->IONum[index]];
        }
        memcpy(fsinfo->IOSize[index], net_fsinfo->IOSize[i], fsinfo->IONum[index]*sizeof(int));

        oldnum = fsinfo->FileNum[index];
        fsinfo->FileNum[index] = net_fsinfo->FileNum[i];
        if (oldnum != fsinfo->FileNum[index])
        {
            delete [] fsinfo->FileMode[index];
            fsinfo->FileMode[index] = new ushort [fsinfo->FileNum[index]];
        }
        memcpy(fsinfo->FileMode[index], net_fsinfo->FileMode[i], fsinfo->FileNum[index]*sizeof(ushort));

        oldnum = SIZE_BIT_TO_BYTE(fsinfo->DirNum[index]);
        fsinfo->DirNum[index] = net_fsinfo->DirNum[i];
        dirmodenum = SIZE_BIT_TO_BYTE(fsinfo->DirNum[index]);
        if (oldnum != dirmodenum)
        {
            delete [] fsinfo->DirMode[index];
            fsinfo->DirMode[index] = new uchar [dirmodenum];
        }
        memcpy(fsinfo->DirMode[index], net_fsinfo->DirMode[i], dirmodenum*sizeof(uchar));

        oldnum = SIZE_BIT_TO_BYTE(fsinfo->SymlinkNum[index]);
        fsinfo->SymlinkNum[index] = net_fsinfo->SymlinkNum[i];
        symlinkmodenum = SIZE_BIT_TO_BYTE(fsinfo->SymlinkNum[index]);
        if (oldnum != symlinkmodenum)
        {
            delete [] fsinfo->SymlinkMode[index];
            fsinfo->SymlinkMode[index] = new uchar [symlinkmodenum];
        }
        memcpy(fsinfo->SymlinkMode[index], net_fsinfo->SymlinkMode[i], symlinkmodenum*sizeof(uchar));
    }
}

void CClientPrimMgrThread::UpdateRefFileSysInfo(int fs_index, CNet_FsInfo *net_fsinfo)
{
    SPrimFileSysInfo *fsinfo = m_fs[fs_index];
    BOOL new_fs = (fsinfo == NULL);

    if (new_fs)
        fsinfo = new SPrimFileSysInfo;

    int i;
    int oldnum, dirmodenum, symlinkmodenum;
    int subdirnum = net_fsinfo->SubDirNum;

    if (new_fs || fsinfo->SubDirNum == 0)
    {
        fsinfo->IONum = new int [subdirnum];
        fsinfo->IOSize = new int * [subdirnum];
        fsinfo->FileNum = new int [subdirnum];
        fsinfo->FileMode = new ushort * [subdirnum];
        fsinfo->DirNum = new int [subdirnum];
        fsinfo->DirMode = new uchar * [subdirnum];
        fsinfo->SymlinkNum = new int [subdirnum];
        fsinfo->SymlinkMode = new uchar * [subdirnum];

        fsinfo->OptDir = new int [subdirnum];

        for (i=0; i<subdirnum; i++)
        {
            fsinfo->IONum[i] = 0;
            fsinfo->IOSize[i] = NULL;
            fsinfo->FileNum[i] = 0;
            fsinfo->FileMode[i] = NULL;
            fsinfo->DirNum[i] = 0;
            fsinfo->DirMode[i] = NULL;
            fsinfo->SymlinkNum[i] = 0;
            fsinfo->SymlinkMode[i] = NULL;
        }

    }

    fsinfo->TotalSize = net_fsinfo->TotalSize;
    fsinfo->TotalFNodeNum = net_fsinfo->TotalFNodeNum;
    fsinfo->SubDirNum = net_fsinfo->SubDirNum;
    fsinfo->TotalIONum = net_fsinfo->TotalIONum;
    fsinfo->TotalFileNum = net_fsinfo->TotalFileNum;
    fsinfo->TotalDirNum = net_fsinfo->TotalDirNum;
    fsinfo->TotalSymlinkNum = net_fsinfo->TotalSymlinkNum;

    for (i=0; i<subdirnum; i++)
    {
        oldnum = fsinfo->IONum[i];
        fsinfo->IONum[i] = net_fsinfo->IONum[i];
        if (oldnum != fsinfo->IONum[i])
        {
            if (fsinfo->IOSize[i] != NULL)
                delete [] fsinfo->IOSize[i];
            fsinfo->IOSize[i] = new int [fsinfo->IONum[i]];
        }
        memcpy(fsinfo->IOSize[i], net_fsinfo->IOSize[i], fsinfo->IONum[i]*sizeof(int));

        fsinfo->TotalFileNum += net_fsinfo->TotalFileNum;
        oldnum = fsinfo->FileNum[i];
        fsinfo->FileNum[i] = net_fsinfo->FileNum[i];
        if (oldnum != fsinfo->FileNum[i])
        {
            if (fsinfo->FileMode[i] != NULL)
                delete [] fsinfo->FileMode[i];
            fsinfo->FileMode[i] = new ushort [fsinfo->FileNum[i]];
        }
        memcpy(fsinfo->FileMode[i], net_fsinfo->FileMode[i], fsinfo->FileNum[i]*sizeof(ushort));

        fsinfo->TotalDirNum += net_fsinfo->TotalDirNum;
        oldnum = SIZE_BIT_TO_BYTE(fsinfo->DirNum[i]);
        fsinfo->DirNum[i] = net_fsinfo->DirNum[i];
        dirmodenum = SIZE_BIT_TO_BYTE(fsinfo->DirNum[i]);
        if (oldnum != dirmodenum)
        {
            if (fsinfo->DirMode[i] != NULL)
                delete [] fsinfo->DirMode[i];
            fsinfo->DirMode[i] = new uchar [dirmodenum];
        }
        memcpy(fsinfo->DirMode[i], net_fsinfo->DirMode[i], dirmodenum*sizeof(uchar));

        fsinfo->TotalSymlinkNum += net_fsinfo->TotalSymlinkNum;
        oldnum = SIZE_BIT_TO_BYTE(fsinfo->SymlinkNum[i]);
        fsinfo->SymlinkNum[i] = net_fsinfo->SymlinkNum[i];
        symlinkmodenum = SIZE_BIT_TO_BYTE(fsinfo->SymlinkNum[i]);
        if (oldnum != symlinkmodenum)
        {
            if (fsinfo->SymlinkMode[i] != NULL)
                delete [] fsinfo->SymlinkMode[i];
            fsinfo->SymlinkMode[i] = new uchar [symlinkmodenum];
        }
        memcpy(fsinfo->SymlinkMode[i], net_fsinfo->SymlinkMode[i], symlinkmodenum*sizeof(uchar));
    }
}

void CClientPrimMgrThread::ShuffleAllFileSystems()
{
    int i, j;
    int usedDirNum;

    int totalLoad = 0;
    for (i=0; i<m_client_num; i++)
        totalLoad += m_load[i];

    for (i=0; i<m_fsnum; i++)
    {
        if (m_fs[i] == NULL)
            continue;

        /*if (m_fs[i]->OptDir != NULL)
            delete [] m_fs[i]->OptDir;
        m_fs[i]->OptDir = new int [m_fs[i]->SubDirNum];*/
        for (j=0; j<m_fs[i]->SubDirNum; j++)
            m_fs[i]->OptDir[j] = j;
        ShuffleArray(m_fs[i]->SubDirNum, m_fs[i]->OptDir);

        usedDirNum = 0;
        for (j=0; j<m_client_num; j++)
        {
            m_fs[i]->OptDirNum[j] = m_fs[i]->SubDirNum * m_load[j] / totalLoad;
            usedDirNum += m_fs[i]->OptDirNum[j];
        }
        /* assign the remainer to first N clients */
        for (j=0; j<m_fs[i]->SubDirNum-usedDirNum; j++)
            m_fs[i]->OptDirNum[j] ++;

        fprintf(stderr, "File system %d:", i);
        for (j=0; j<m_client_num; j++)
            fprintf(stderr, " %d", m_fs[i]->OptDirNum[j]);
        fprintf(stderr, "\n");
    }
}

BOOL CClientPrimMgrThread::RecvFileSysInfoFromRM(CSocketClient *socket)
{
    CPacket packet;
    char *buffer;
    int size;

    /* receive the file system number */
    printf("Receiving file system number ...\n");
    if (! packet.Receive(socket))
        return FALSE;
    buffer = packet.GetData(&size);
    printf("Got %d bytes\n", size);
    CNet_CommNum net_fsnum;
    net_fsnum.xdt_trans(XDTOP_GET, buffer, size);
    m_fsnum = net_fsnum.num;

    int i, j;
    for (i=0; i<m_fsnum; i++)
        m_fs[i] = NULL;

    /* receive file system information one by one */
    CNet_FsLocation net_fslocation;
    CNet_FsInfo net_fsinfo;
    int subdirnum, dirmodenum, symlinkmodenum;
    for (i=0; i<m_fsnum; i++)
    {
        printf("Receiving file system (%d of %d) information ...\n", i, m_fsnum);

        if (! packet.Receive(socket))
            return FALSE;
        buffer = packet.GetData(&size);
        printf("Got file system lcoation (%d bytes)\n", size);
        net_fslocation.xdt_trans(XDTOP_GET, buffer, size);

        if (! packet.Receive(socket))
            return FALSE;
        buffer = packet.GetData(&size);
        printf("Got file system info (%d bytes)\n", size);
        net_fsinfo.xdt_trans(XDTOP_GET, buffer, size);

        m_fs[i] = new SPrimFileSysInfo;

        strcpy(m_fs[i]->ServerName, net_fslocation.ServerName);
        strcpy(m_fs[i]->FsName, net_fslocation.FsName);
        strcpy(m_fs[i]->ServerNetIF, net_fslocation.ServerNetIF);
        strcpy(m_fs[i]->ServerPath, net_fslocation.ServerPath);
        strcpy(m_fs[i]->ShareName, net_fslocation.ShareName);
        m_fs[i]->FsSize = net_fslocation.FsSize;

        m_fs[i]->TotalSize = net_fsinfo.TotalSize;
        m_fs[i]->TotalFNodeNum = net_fsinfo.TotalFNodeNum;
        m_fs[i]->SubDirNum = net_fsinfo.SubDirNum;
        m_fs[i]->TotalIONum = net_fsinfo.TotalIONum;
        m_fs[i]->TotalFileNum = net_fsinfo.TotalFileNum;
        m_fs[i]->TotalDirNum = net_fsinfo.TotalDirNum;
        m_fs[i]->TotalSymlinkNum = net_fsinfo.TotalSymlinkNum;

        subdirnum = net_fsinfo.SubDirNum;

        m_fs[i]->IONum = new int [subdirnum];
        memcpy(m_fs[i]->IONum, net_fsinfo.IONum, subdirnum * sizeof(int));

        m_fs[i]->IOSize = new int * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            m_fs[i]->IOSize[j] = new int [m_fs[i]->IONum[j]];
            memcpy(m_fs[i]->IOSize[j], net_fsinfo.IOSize[j], m_fs[i]->IONum[j]*sizeof(int));
        }

        m_fs[i]->FileNum = new int [subdirnum];
        memcpy(m_fs[i]->FileNum, net_fsinfo.FileNum, subdirnum * sizeof(int));

        m_fs[i]->FileMode = new ushort * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            m_fs[i]->FileMode[j] = new ushort [m_fs[i]->FileNum[j]];
            memcpy(m_fs[i]->FileMode[j], net_fsinfo.FileMode[j], m_fs[i]->FileNum[j] * sizeof(ushort));
        }

        m_fs[i]->DirNum = new int [subdirnum];
        memcpy(m_fs[i]->DirNum, net_fsinfo.DirNum, subdirnum * sizeof(int));

        m_fs[i]->DirMode = new uchar * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            dirmodenum = SIZE_BIT_TO_BYTE(m_fs[i]->DirNum[j]);
            m_fs[i]->DirMode[j] = new uchar [dirmodenum];
            memcpy(m_fs[i]->DirMode[j], net_fsinfo.DirMode[j], dirmodenum * sizeof(uchar));
        }

        m_fs[i]->SymlinkNum = new int [subdirnum];
        memcpy(m_fs[i]->SymlinkNum, net_fsinfo.SymlinkNum, subdirnum * sizeof(int));

        m_fs[i]->SymlinkMode = new uchar * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            symlinkmodenum = SIZE_BIT_TO_BYTE(m_fs[i]->SymlinkNum[j]);
            m_fs[i]->SymlinkMode[j] = new uchar [symlinkmodenum];
            memcpy(m_fs[i]->SymlinkMode[j], net_fsinfo.SymlinkMode[j], symlinkmodenum * sizeof(uchar));
        }

        m_fs[i]->OptDir = new int [m_fs[i]->SubDirNum];
    }

    return TRUE;
}
    

BOOL CClientPrimMgrThread::RecvAckFromClient(CSocketClient *socket)
{
    CPacket packet;
    char *buffer;
    int size;

    /* receive client id first */
    if (! packet.Receive(socket))
        return FALSE;
    buffer = packet.GetData(&size);
    CNet_ClientID net_clientid;
    net_clientid.xdt_trans(XDTOP_GET, buffer, size);
    int client_id = net_clientid.id;

    /* receive ack number */
    if (! packet.Receive(socket))
        return FALSE;
    CNet_CommNum net_ack;
    buffer = packet.GetData(&size);
    net_ack.xdt_trans(XDTOP_GET, buffer, size);
    int ack = net_ack.num;

    AddLcReady(client_id, ack);

    return TRUE;
}


BOOL CClientPrimMgrThread::RecvFileSysInfoFromClient(CSocketClient *socket)
{
    CPacket packet;
    char *buffer;
    int size;

    /* receive client id first */
    if (! packet.Receive(socket))
        return FALSE;

    buffer = packet.GetData(&size);
    CNet_ClientID net_clientid;
    net_clientid.xdt_trans(XDTOP_GET, buffer, size);
    int client_id = net_clientid.id;

    /* receive the net_fsinfo data */
    CNet_FsSysInfo net_fsinfo;
    for (int i=0; i<m_fsnum; i++)
    {
        if (! packet.Receive(socket))
            return FALSE;
        buffer = packet.GetData(&size);
        net_fsinfo.xdt_trans(XDTOP_GET, buffer, size);
        UpdateFileSysInfo(client_id, i, &net_fsinfo);
    }

    AddLcReady(client_id, 0);

    return TRUE;
}


BOOL CClientPrimMgrThread::RecvRefFileSysInfoFromClient(CSocketClient *socket)
{
    CPacket packet;
    char *buffer;
    int size;

    /* receive client id first, use client id as file system index */
    if (! packet.Receive(socket))
        return FALSE;

    buffer = packet.GetData(&size);
    CNet_ClientID net_clientid;
    net_clientid.xdt_trans(XDTOP_GET, buffer, size);
    int fs_index = net_clientid.id;

    /* receive the net_fsinfo data */
    CNet_FsInfo net_fsinfo;
    if (! packet.Receive(socket))
        return FALSE;
    buffer = packet.GetData(&size);
    net_fsinfo.xdt_trans(XDTOP_GET, buffer, size);
    fprintf(stderr, "Got message: %d IO %d NonIO %d Dir\n", net_fsinfo.TotalIONum, net_fsinfo.TotalFileNum, net_fsinfo.TotalDirNum);
    UpdateRefFileSysInfo(fs_index, &net_fsinfo);

    return TRUE;
}


BOOL CClientPrimMgrThread::SendLoadRunResponseToRM(CSocketClient *socket, BOOL succ, char *error_msg)
{
    CPacket packet;
    char *buffer;
    int size;

    CNet_StartLoadOnPrimResponse net_resp;
    net_resp.succ = succ;
    if (error_msg != NULL)
        strcpy(net_resp.error_msg, error_msg);
    else
        strcpy(net_resp.error_msg, succ ? "Successful" : "General error");
    net_resp.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(socket))
        return FALSE;

    return TRUE;
}


BOOL CClientPrimMgrThread::SendLoadUpdateToRM(int status, int usedtime, char *error_msg)
{
    CSocketClient socket;
    CPacket packet;
    char *buffer;
    int size;

    if (! socket.ConnectToServer(m_rmip, REQ_SERVER_PORT))
    {
        fprintf(stderr, "Cannot connect to resource manager server (%s).\n", m_rmname);
        return FALSE;
    }

    /* message type */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_FAUCET_LOADUPDATE;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;

    /* load update */
    CNet_LoadUpdate net_update;
    net_update.id = m_load_id;
    net_update.status = status;
    if (status == 99 && error_msg != NULL)
        strcpy(net_update.error_msg, error_msg);
    else
        strcpy(net_update.error_msg, "");
    net_update.usedtime = usedtime;
    net_update.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
        return FALSE;
//    fprintf(stderr, "%d bytes (load update) sent\n", size);

    /* for load done (finished, paused, stopped), need send more information (file system information) */
    if (status == 0 || status == 1 || status == 2)
    {
        /* file system number */
        CNet_CommNum net_fsnum;
        net_fsnum.num = m_fsnum;
        net_fsnum.xdt_trans(XDTOP_PUT, buffer, size);
        packet.SetData(buffer, size);
        if (! packet.Send(&socket))
            return FALSE;
//        fprintf(stderr, "%d bytes (file system number) sent\n", size);

        /* file system information */
        CNet_FsInfo net_fsinfo;
        for (int i=0; i<m_fsnum; i++)
        {
            net_fsinfo.TotalSize        = m_fs[i]->TotalSize;
            net_fsinfo.TotalFNodeNum    = m_fs[i]->TotalFNodeNum;
            net_fsinfo.SubDirNum        = m_fs[i]->SubDirNum;
            net_fsinfo.TotalIONum       = m_fs[i]->TotalIONum;
            net_fsinfo.TotalFileNum     = m_fs[i]->TotalFileNum;
            net_fsinfo.TotalDirNum      = m_fs[i]->TotalDirNum;
            net_fsinfo.TotalSymlinkNum  = m_fs[i]->TotalSymlinkNum;
            net_fsinfo.IONum            = m_fs[i]->IONum;
            net_fsinfo.IOSize           = m_fs[i]->IOSize;
            net_fsinfo.FileNum          = m_fs[i]->FileNum;
            net_fsinfo.FileMode         = m_fs[i]->FileMode;
            net_fsinfo.DirNum           = m_fs[i]->DirNum;
            net_fsinfo.DirMode          = m_fs[i]->DirMode;
            net_fsinfo.SymlinkNum       = m_fs[i]->SymlinkNum;
            net_fsinfo.SymlinkMode      = m_fs[i]->SymlinkMode;

            net_fsinfo.xdt_trans(XDTOP_PUT, buffer, size);
            packet.SetData(buffer, size);
            if (! packet.Send(&socket))
                return FALSE;
//            fprintf(stderr, "%d bytes (file system information) sent\n", size);
        }
    }

    return TRUE;
}


BOOL CClientPrimMgrThread::SendStartLoadOnClients(CSocketClient *sock_to_clients, char *adv_buffer, int adv_size, char *error_msg)
{
    CPacket packet1, packet2, packet3;
    char *buffer1, *buffer2, *buffer3;
    int size1, size2, size3;

    /* message type */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_START_LOAD;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer1, size1);
    packet1.SetData(buffer1, size1);

    /* advanced setting */
    buffer3 = adv_buffer;
    size3 = adv_size;
    packet3.SetData(buffer3, size3);

    /* start load message */
    CNet_StartLoadOnClient net_startload;

    int i;
    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        /* 1. send message type */
        if (! packet1.Send(&sock_to_clients[i]))
        {
            sprintf(error_msg, "Can't send message type to client %s", m_clients[i]);
            return FALSE;
        }

        /* 2. send load parameters */
        net_startload.load_id = m_load_id;
        net_startload.client_id = i;
        strcpy(net_startload.prim_name, m_myname);
        strcpy(net_startload.prim_ip, m_myip);
        strcpy(net_startload.load_name, m_clients[i]);
        strcpy(net_startload.load_ip, m_client_ips[i]);
        net_startload.opnum = m_load[i];
        net_startload.stress_server = m_stress_server;
        net_startload.run_time = m_run_time;
        net_startload.xdt_trans(XDTOP_PUT, buffer2, size2);
        packet2.SetData(buffer2, size2);
        if (! packet2.Send(&sock_to_clients[i]))
        {
            sprintf(error_msg, "Can't send load parameters to client %s", m_clients[i]);
            return FALSE;
        }

        /* 3. send advanced setting */
        if (! packet3.Send(&sock_to_clients[i]))
        {
            sprintf(error_msg, "Can't send load advanced setting to client %s", m_clients[i]);
            return FALSE;
        }
    }

    return TRUE;
}


BOOL CClientPrimMgrThread::SendFSBaseInfoToClients(CSocketClient *sock_to_clients, char *error_msg)
{
    CPacket packet4, packet5;
    char *buffer4, **buffer5;
    int size4, *size5;

    /* file system number */
    CNet_CommNum net_fsnum;
    net_fsnum.num = m_fsnum;
    net_fsnum.xdt_trans(XDTOP_PUT, buffer4, size4);
    packet4.SetData(buffer4, size4);

    /* file system base information */
    CNet_FsSysBase *net_fsbase;
    net_fsbase = new CNet_FsSysBase [m_fsnum];
    buffer5 = new char * [m_fsnum];
    size5 = new int [m_fsnum];

    int i, j;
    for (i=0; i<m_fsnum; i++)
    {
        strcpy(net_fsbase[i].ServerName, m_fs[i]->ServerName);
        strcpy(net_fsbase[i].ServerNetIF, m_fs[i]->ServerNetIF);
        strcpy(net_fsbase[i].ServerPath, m_fs[i]->ServerPath);
        strcpy(net_fsbase[i].ShareName, m_fs[i]->ShareName);

        net_fsbase[i].TotalSize         = m_fs[i]->TotalSize;
        net_fsbase[i].TotalFNodeNum     = m_fs[i]->TotalFNodeNum;
        net_fsbase[i].SubDirNum         = m_fs[i]->SubDirNum;
        net_fsbase[i].TotalIONum        = m_fs[i]->TotalIONum;
        net_fsbase[i].TotalFileNum      = m_fs[i]->TotalFileNum;
        net_fsbase[i].TotalDirNum       = m_fs[i]->TotalDirNum;
        net_fsbase[i].TotalSymlinkNum   = m_fs[i]->TotalSymlinkNum;

        net_fsbase[i].xdt_trans(XDTOP_PUT, buffer5[i], size5[i]);
    }

    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        /* 4. send file system number */
        fprintf(stderr, "Send file system number (%d) (%d bytes) to client (%s)\n", m_fsnum, size4, m_clients[i]);
        if (! packet4.Send(&sock_to_clients[i]))
        {
            delete [] buffer5;
            delete [] size5;
            delete [] net_fsbase;
            sprintf(error_msg, "Can't send file system number to client %s", m_clients[i]);
            return FALSE;
        }

        /* 5. send file system base information */
        for (j=0; j<m_fsnum; j++)
        {
            fprintf(stderr, "Send file system (%d of %d) base information (%d bytes) to client (%s)\n", j+1, m_fsnum, size5[j], m_clients[i]);
            packet5.SetData(buffer5[j], size5[j]);
            if (! packet5.Send(&sock_to_clients[i]))
            {
                delete [] buffer5;
                delete [] size5;
                delete [] net_fsbase;
                sprintf(error_msg, "Can't send file system (%d of %d) basic information to client %s", j+1, m_fsnum, m_clients[i]);
                return FALSE;
            }
        }
    }

    delete [] buffer5;
    delete [] size5;
    delete [] net_fsbase;

    return TRUE;
}


BOOL CClientPrimMgrThread::SendGoAheadToClients(CSocketClient *sock_to_clients, char *error_msg)
{
    CPacket packet6;
    char *buffer6;
    int size6;

    /* go ahead signal with magic number */
    CNet_CommNum net_signal;
    net_signal.num = 0x4321cdef;
    net_signal.xdt_trans(XDTOP_PUT, buffer6, size6);
    packet6.SetData(buffer6, size6);

    for (int i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        /* 6. send go ahead signal */
        if (! packet6.Send(&sock_to_clients[i]))
        {
            sprintf(error_msg, "Can't send go ahead signal to client %s", m_clients[i]);
            return FALSE;
        }
    }

    return TRUE;
}

BOOL CClientPrimMgrThread::SendFileSysInfoToClients(CSocketClient *sock_to_clients, char *error_msg)
{
    int i, j, k, index, start_index, optdirnum, dirmodenum, symlinkmodenum;
    char ctime_buf[32];

    int prev_OptDirIndex[MAX_FSNUM];
    for (i=0; i<m_fsnum; i++)
        prev_OptDirIndex[i] = 0;

    CPacket packet7;
    char *buffer7;
    int size7;

    CNet_FsSysInfo net_fssysinfo;

    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        /* file system information one by one */
        for (j=0; j<m_fsnum; j++)
        {
            start_index = prev_OptDirIndex[j];
            optdirnum = m_fs[j]->OptDirNum[i];
            prev_OptDirIndex[j] += optdirnum;

            /* fill OptDirNum */
            net_fssysinfo.OptDirNum = m_fs[j]->OptDirNum[i];

            /* fill TotalFileNum, TotalDirNum, TotalSymlinkNum;   load clients will calculate these numbers themselves */
            net_fssysinfo.TotalFileNum = 0;
            net_fssysinfo.TotalDirNum = 0;
            net_fssysinfo.TotalSymlinkNum = 0;

            /* fill OptDir */
            net_fssysinfo.OptDir = new int [optdirnum];
            memcpy(net_fssysinfo.OptDir, &m_fs[j]->OptDir[start_index], optdirnum*sizeof(int));

            /* fill IONum */
            net_fssysinfo.IONum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.IONum[k] = m_fs[j]->IONum[index];
            }

            /* fill IOSize */
            net_fssysinfo.IOSize = new int * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.IOSize[k] = new int [m_fs[j]->IONum[index]];
                memcpy(net_fssysinfo.IOSize[k], m_fs[j]->IOSize[index], m_fs[j]->IONum[index]*sizeof(int));
            }

            /* fill FileNum */
            net_fssysinfo.FileNum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.FileNum[k] = m_fs[j]->FileNum[index];
            }

            /* fill FileMode */
            net_fssysinfo.FileMode = new ushort * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.FileMode[k] = new ushort [m_fs[j]->FileNum[index]];
                memcpy(net_fssysinfo.FileMode[k], m_fs[j]->FileMode[index], m_fs[j]->FileNum[index]*sizeof(ushort));
            }

            /* fill DirNum */
            net_fssysinfo.DirNum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.DirNum[k] = m_fs[j]->DirNum[index];
            }

            /* fill DirMode */
            net_fssysinfo.DirMode = new uchar * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                dirmodenum = SIZE_BIT_TO_BYTE(m_fs[j]->DirNum[index]);
                net_fssysinfo.DirMode[k] = new uchar [dirmodenum];
                memcpy(net_fssysinfo.DirMode[k], m_fs[j]->DirMode[index], dirmodenum*sizeof(uchar));
            }

            /* fill SymlinkNum */
            net_fssysinfo.SymlinkNum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.SymlinkNum[k] = m_fs[j]->SymlinkNum[index];
            }

            /* fill SymlinkMode */
            net_fssysinfo.SymlinkMode = new uchar * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                symlinkmodenum = SIZE_BIT_TO_BYTE(m_fs[j]->SymlinkNum[index]);
                net_fssysinfo.SymlinkMode[k] = new uchar [symlinkmodenum];
                memcpy(net_fssysinfo.SymlinkMode[k], m_fs[j]->SymlinkMode[index], symlinkmodenum*sizeof(uchar));
            }

            /* the data is ready, now buffer it and send it */
            net_fssysinfo.xdt_trans(XDTOP_PUT, buffer7, size7);

            /* delete the temp data */
            for (k=0; k<optdirnum; k++)
            {
                delete [] net_fssysinfo.IOSize[k];
                delete [] net_fssysinfo.FileMode[k];
                delete [] net_fssysinfo.DirMode[k];
                delete [] net_fssysinfo.SymlinkMode[k];
            }
            delete [] net_fssysinfo.OptDir;
            delete [] net_fssysinfo.IONum;
            delete [] net_fssysinfo.IOSize;
            delete [] net_fssysinfo.FileNum;
            delete [] net_fssysinfo.FileMode;
            delete [] net_fssysinfo.DirNum;
            delete [] net_fssysinfo.DirMode;
            delete [] net_fssysinfo.SymlinkNum;
            delete [] net_fssysinfo.SymlinkMode;

            /* send the data in the buffer */
            printf("%s: Send file system (%d of %d) information (%d bytes) to client (%s)\n", Safe_ctime(ctime_buf), j+1, m_fsnum, size7, m_clients[i]);
            fflush(stdout);
            packet7.SetData(buffer7, size7);
            if (! packet7.Send(&sock_to_clients[i]))
            {
                sprintf(error_msg, "Can't send file system (%s) information to client (%s)", m_fs[j]->FsName, m_clients[i]);
                return FALSE;
            }
        }
    }

    return TRUE;
}

BOOL CClientPrimMgrThread::SendBreakMessageToClients(int status, char *error_msg)
{
    CPacket packet1, packet2, packet3;
    char *buffer1, *buffer2, *buffer3;
    int size1, size2, size3;

    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_BREAK;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer1, size1);
    packet1.SetData(buffer1, size1);

    CNet_LoadID net_loadid;
    net_loadid.id = m_load_id;
    net_loadid.xdt_trans(XDTOP_PUT, buffer2, size2);
    packet2.SetData(buffer2, size2);

    CNet_BreakMsg net_breakmsg;
    net_breakmsg.reason = status;
    net_breakmsg.xdt_trans(XDTOP_PUT, buffer3, size3);
    packet3.SetData(buffer3, size3);

    CSocketClient socket;
    BOOL connected[MAX_CLIENT_NUM], sent[MAX_CLIENT_NUM];
    int i;

    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        connected[i] = socket.ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT);
        if (! connected[i])     /* don't give up, continue to send to next client */
            continue;

        /* message type, load id and break message */
        if (packet1.Send(&socket) && packet2.Send(&socket) && packet3.Send(&socket))
            sent[i] = TRUE;
        else
            sent[i] = FALSE;
    }

    BOOL allConnected = TRUE;
    BOOL allSent = TRUE;
    char error_msg1[MAX_MSGLEN], error_msg2[MAX_MSGLEN];

    strcpy(error_msg1, "Cannot connect to load client (");
    strcpy(error_msg2, "Cannot send data to load client (");
    for (i=0; i<m_client_num; i++)
    {
        if (! connected[i])
        {
            strcat(error_msg1, " ");
            strcat(error_msg1, m_clients[i]);
            allConnected = FALSE;
        }

        if (! sent[i])
        {
            strcat(error_msg2, " ");
            strcat(error_msg2, m_clients[i]);
            allSent = FALSE;
        }
    }
    strcat(error_msg1, " )");
    strcat(error_msg2, " )");

    if (allConnected && allSent)
        return TRUE;
    else
    {
        strcpy(error_msg, "");
        if (! allConnected)
        {
            strcat(error_msg, error_msg1);
            strcat(error_msg, " ");
        }
        if (! allSent)
            strcat(error_msg, error_msg2);
        return FALSE;
    }
}

BOOL CClientPrimMgrThread::SendGoAheadToClients(char *error_msg)
{
    CPacket packet1, packet2, packet3;
    char *buffer1, *buffer2, *buffer3;
    int size1, size2, size3;

    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_GOAHEAD;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer1, size1);
    packet1.SetData(buffer1, size1);

    CNet_LoadID net_loadid;
    net_loadid.id = m_load_id;
    net_loadid.xdt_trans(XDTOP_PUT, buffer2, size2);
    packet2.SetData(buffer2, size2);

    CNet_CommNum net_signal;
    net_signal.num = 0x4321cdef;
    net_signal.xdt_trans(XDTOP_PUT, buffer3, size3);
    packet3.SetData(buffer3, size3);

    CSocketClient socket[MAX_CLIENT_NUM];
    int i;

    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        if (! socket[i].ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT))
        {
            sprintf(error_msg, "Cannot connect to load client (%s)", m_clients[i]);
            return FALSE;
        }
    }

    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        /* message type, load id and break message */
        if (!packet1.Send(&socket[i]) || !packet2.Send(&socket[i]) || !packet3.Send(&socket[i]))
        {
            sprintf(error_msg, "Can't send go ahead signal to client (%s)", m_clients[i]);
            return FALSE;
        }
    }

    return TRUE;
}

BOOL CClientPrimMgrThread::SendFileSysInfoToClients(char *error_msg)
{
    int i, j, k, index, start_index, optdirnum, dirmodenum, symlinkmodenum;
    char ctime_buf[32];

    int prev_OptDirIndex[MAX_FSNUM];
    for (i=0; i<m_fsnum; i++)
        prev_OptDirIndex[i] = 0;

    CNet_MsgType net_msgtype;
    char *buffer1;
    int size1;
    net_msgtype.type = NET_MSGTYPE_PUMPER_FSINFO;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer1, size1);

    CNet_LoadID net_loadid;
    char *buffer2;
    int size2;
    net_loadid.id = m_load_id;
    net_loadid.xdt_trans(XDTOP_PUT, buffer2, size2);

    CSocketClient socket;
    CPacket packet;
    CNet_FsSysInfo net_fssysinfo;
    char *buffer3;
    int size3;

    for (i=0; i<m_client_num; i++)
    {
        /* send to local machine */
        //if (strcmp(m_myname, m_clients[i]) == 0)
        {
        }

        if (! socket.ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT))
        {
            fprintf(stderr, "Cannot connect to load client %s.\n", m_clients[i]);
            return FALSE;
        }

        /* message type */
        packet.SetData(buffer1, size1);
        if (! packet.Send(&socket))
            return FALSE;

        /* load id */
        packet.SetData(buffer2, size2);
        if (! packet.Send(&socket))
            return FALSE;

        /* file system information one by one */
        for (j=0; j<m_fsnum; j++)
        {
            start_index = prev_OptDirIndex[j];
            optdirnum = m_fs[j]->OptDirNum[i];
            prev_OptDirIndex[j] += optdirnum;

            /* fill OptDirNum */
            net_fssysinfo.OptDirNum = m_fs[j]->OptDirNum[i];

            /* fill TotalFileNum, TotalDirNum, TotalSymlinkNum;   load clients will calculate these numbers themselves */
            net_fssysinfo.TotalFileNum = 0;
            net_fssysinfo.TotalDirNum = 0;
            net_fssysinfo.TotalSymlinkNum = 0;

            /* fill OptDir */
            net_fssysinfo.OptDir = new int [optdirnum];
            memcpy(net_fssysinfo.OptDir, &m_fs[j]->OptDir[start_index], optdirnum*sizeof(int));

            /* fill IONum */
            net_fssysinfo.IONum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.IONum[k] = m_fs[j]->IONum[index];
            }

            /* fill IOSize */
            net_fssysinfo.IOSize = new int * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.IOSize[k] = new int [m_fs[j]->IONum[index]];
                memcpy(net_fssysinfo.IOSize[k], m_fs[j]->IOSize[index], m_fs[j]->IONum[index]*sizeof(int));
            }

            /* fill FileNum */
            net_fssysinfo.FileNum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.FileNum[k] = m_fs[j]->FileNum[index];
            }

            /* fill FileMode */
            net_fssysinfo.FileMode = new ushort * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.FileMode[k] = new ushort [m_fs[j]->FileNum[index]];
                memcpy(net_fssysinfo.FileMode[k], m_fs[j]->FileMode[index], m_fs[j]->FileNum[index]*sizeof(ushort));
            }

            /* fill DirNum */
            net_fssysinfo.DirNum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.DirNum[k] = m_fs[j]->DirNum[index];
            }

            /* fill DirMode */
            net_fssysinfo.DirMode = new uchar * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                dirmodenum = SIZE_BIT_TO_BYTE(m_fs[j]->DirNum[index]);
                net_fssysinfo.DirMode[k] = new uchar [dirmodenum];
                memcpy(net_fssysinfo.DirMode[k], m_fs[j]->DirMode[index], dirmodenum*sizeof(uchar));
            }

            /* fill SymlinkNum */
            net_fssysinfo.SymlinkNum = new int [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                net_fssysinfo.SymlinkNum[k] = m_fs[j]->SymlinkNum[index];
            }

            /* fill SymlinkMode */
            net_fssysinfo.SymlinkMode = new uchar * [optdirnum];
            for (k=0; k<optdirnum; k++)
            {
                index = m_fs[j]->OptDir[start_index + k];
                symlinkmodenum = SIZE_BIT_TO_BYTE(m_fs[j]->SymlinkNum[index]);
                net_fssysinfo.SymlinkMode[k] = new uchar [symlinkmodenum];
                memcpy(net_fssysinfo.SymlinkMode[k], m_fs[j]->SymlinkMode[index], symlinkmodenum*sizeof(uchar));
            }

            /* the data is ready, now buffer it and send it */
            net_fssysinfo.xdt_trans(XDTOP_PUT, buffer3, size3);

            /* delete the temp data */
            for (k=0; k<optdirnum; k++)
            {
                delete [] net_fssysinfo.IOSize[k];
                delete [] net_fssysinfo.FileMode[k];
                delete [] net_fssysinfo.DirMode[k];
                delete [] net_fssysinfo.SymlinkMode[k];
            }
            delete [] net_fssysinfo.OptDir;
            delete [] net_fssysinfo.IONum;
            delete [] net_fssysinfo.IOSize;
            delete [] net_fssysinfo.FileNum;
            delete [] net_fssysinfo.FileMode;
            delete [] net_fssysinfo.DirNum;
            delete [] net_fssysinfo.DirMode;
            delete [] net_fssysinfo.SymlinkNum;
            delete [] net_fssysinfo.SymlinkMode;

            /* send the data in the buffer */
            printf("%s: Send file system (%d of %d) information (%d bytes) to client (%s)\n", Safe_ctime(ctime_buf), j+1, m_fsnum, size3, m_clients[i]);
            fflush(stdout);
            packet.SetData(buffer3, size3);
            if (! packet.Send(&socket))
                return FALSE;
        }
    }

    return TRUE;
}

void CClientPrimMgrThread::SendQuitMessageClients()
{
    char ctime_buf[32];

    CNet_MsgType net_msgtype;
    CPacket packet;
    char *buffer;
    int size;
    net_msgtype.type = NET_MSGTYPE_EXIT_SERVER;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);

    CSocketClient sock;
    for (int i=0; i<m_client_num; i++)
    {
        if (! sock.ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT))
        {
            printf("%s: Cannot connect to load client %d: %s (%s).\n", Safe_ctime(ctime_buf), i+1, m_clients[i], m_client_ips[i]);
            fflush(stdout);
        }
        else if (!packet.Send(&sock))
        {
            printf("%s: Error to send message to load client %d: %s (%s).\n", Safe_ctime(ctime_buf), i+1, m_clients[i], m_client_ips[i]);
            fflush(stdout);
        }
    }
}

void CClientPrimMgrThread::ResetFileSysTotalNum()
{
    for (int i=0; i<m_fsnum; i++)
    {
        m_fs[i]->TotalFileNum = 0;
        m_fs[i]->TotalDirNum = 0;
        m_fs[i]->TotalSymlinkNum = 0;
    }
}

/* check whether to refresh file system */
BOOL CClientPrimMgrThread::CheckForRefresh(int index)
{

//return TRUE;

    int i, j;
    double percent;
    char ctime_buf[32];

    /* fprintf(stderr, "Check for refreshing ... \n"); */

    /* 1. check summary of IO file size */
    int TotalFileSize = 0;        /* in KBytes */
    int subdirnum = m_fs[index]->SubDirNum;
    int totalsubsize;
    for (i=0; i<subdirnum; i++)
    {
        totalsubsize = 0;         /* in Bytes */
        for (j=0; j<m_fs[index]->IONum[i]; j++)
            totalsubsize += m_fs[index]->IOSize[i][j];
        TotalFileSize += totalsubsize / KB;
    }
    fprintf(stderr, "    TotalUsedFileSize = %d(KB),  TotalFileSize = %d(KB)\n", TotalFileSize, m_fs[index]->TotalSize);

    percent = (double) TotalFileSize / (double) m_fs[index]->TotalSize;
    if (percent > 0.8)
    {
        printf("%s: Size: %d / %d > 0.8\n", Safe_ctime(ctime_buf), TotalFileSize, m_fs[index]->TotalSize);
        fflush(stdout);
        return TRUE;
    }


    /* 2. check inode number */
    int TotalUsedFNodeNum = m_fs[index]->TotalFileNum + m_fs[index]->TotalDirNum + m_fs[index]->TotalSymlinkNum;
    fprintf(stderr, "    TotalUsedFNodeNum = %d,  TotalFNodeNum = %d\n", TotalUsedFNodeNum, m_fs[index]->TotalFNodeNum);
    percent = (double) TotalUsedFNodeNum / (double) m_fs[index]->TotalFNodeNum;
    if (percent > 0.8)
    {
        printf("%s: Inode: (%d + %d + %d) / %d > 0.8\n", Safe_ctime(ctime_buf), m_fs[index]->TotalFileNum, m_fs[index]->TotalDirNum,
                                                         m_fs[index]->TotalSymlinkNum, m_fs[index]->TotalFNodeNum);
        fflush(stdout);
        return TRUE;
    }

    /* 3. check max file,dir,symlink name number */
    int max_file, max_dir, max_symlink;
    max_file = max_dir = max_symlink = -1;
    for (i=0; i<subdirnum; i++)
    {
        /* get actual file number for each directory */
        for (j=m_fs[index]->FileNum[i]-1; j>=0; j--)
        {
            if ((m_fs[index]->FileMode[i][j] & 0x200) == 0)
                break;
        }
        /* actual file number is j+1 */
        if (j + 1 >= max_file)
            max_file = j + 1;

        /* get actual dir number for each directory */
        for (j=m_fs[index]->DirNum[i]-1; j>=0; j--)
        {
            if (((m_fs[index]->DirMode[i][j/8] >> (j % 8)) & 0x01) != 0)
                break;
        }
        /* actual dir number is j+1 */
        if (j + 1 >= max_dir)
            max_dir = j + 1;

        /* get actual symlink number for each directory */
        for (j=m_fs[index]->SymlinkNum[i]-1; j>=0; j--)
        {
            if (((m_fs[index]->SymlinkMode[i][j/8] >> (j % 8)) & 0x01) != 0)
                break;
        }
        /* actual symlink number is j+1 */
        if (j + 1 >= max_symlink)
            max_symlink = j + 1;
    }

    int max_subdirnum = -1;
    for (i=0; i<m_client_num; i++)
    {
        if (m_fs[index]->OptDirNum[i] > max_subdirnum)
            max_subdirnum = m_fs[index]->OptDirNum[i];
    }
    max_file *= max_subdirnum;
    max_dir *= max_subdirnum;
    max_symlink *= max_subdirnum;

    fprintf(stderr, "    max_file=%d(%d),   max_dir=%d(%d),   max_symlink=%d(%d)\n", max_file, MAX_FILEINDEX * 4 / 5, 
                                                                            max_dir, MAX_DIRINDEX * 4 / 5,
                                                                            max_symlink, MAX_SYMLINKINDEX * 4 / 5);
    if (max_file > MAX_FILEINDEX * 4 / 5)
    {
        printf("%s: max_file(%d) > MAX_FILEINDEX*0.8 (%d)\n", Safe_ctime(ctime_buf), max_file, MAX_FILEINDEX * 0.8);
        fflush(stdout);
        return TRUE;
    }
    if (max_dir > MAX_DIRINDEX * 4 / 5)
    {
        printf("%s: max_dir(%d) > MAX_DIRINDEX*0.8 (%d)\n", Safe_ctime(ctime_buf), max_dir, MAX_DIRINDEX * 0.8);
        fflush(stdout);
        return TRUE;
    }
    if (max_symlink > MAX_SYMLINKINDEX * 4 / 5)
    {
        printf("%s: max_symlink(%d) > MAX_SYMLINKINDEX*0.8 (%d)\n", Safe_ctime(ctime_buf), max_symlink, MAX_SYMLINKINDEX * 0.8);
        fflush(stdout);
        return TRUE;
    }

    return FALSE;
}

/* using load client to refresh file systems */
/* return value: 0:Error, 1:All finished 2:paused 3:stopped */
int CClientPrimMgrThread::RefeshFileSystems(int ref_fs_num, int ref_fs_index[])
{
    char ctime_buf[32];
    int not_finish_num = ref_fs_num;
    int ref_num;
    int i, index, sig_code;

    CNet_MsgType net_msgtype;
    CNet_MsgInstFileSys net_msg;    /* for start refreshing */
    CNet_LoadID net_loadid;         /* for stop refreshing */
    CSocketClient socket;
    CPacket packet;
    char *buffer;
    int size;

    BOOL succ = TRUE;

    while (not_finish_num > 0)
    {
        ref_num = (m_client_num <= not_finish_num) ? m_client_num : not_finish_num;
        SetRefClientNum(ref_num);
        ResetRefReady();

        for (i=0; i<ref_num; i++)
        {
            if (! socket.ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT))
            {
                fprintf(stderr, "Cannot connect to load client %s.\n", m_clients[i]);
                succ = FALSE;
                continue;
            }

            /* message type */
            net_msgtype.type = NET_MSGTYPE_PUMPER_START_REFFS;
            net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
            packet.SetData(buffer, size);
            if (! packet.Send(&socket))
            {
                succ = FALSE;
                continue;
            }

            index = ref_fs_index[ref_fs_num - not_finish_num + i];

            net_msg.job_id = m_load_id;     /* use job_id to store load_id */
            net_msg.client_id = index;      /* use client_id to store index of file system */
            strcpy(net_msg.HostName, m_myname);
            strcpy(net_msg.HostIp, m_myip);
            strcpy(net_msg.ServerName, m_fs[index]->ServerName);
            strcpy(net_msg.FSName, m_fs[index]->FsName);
            strcpy(net_msg.ServerNetIF, m_fs[index]->ServerNetIF);
            strcpy(net_msg.ServerPath, m_fs[index]->ServerPath);
            strcpy(net_msg.ServerShare, m_fs[index]->ShareName);
            net_msg.FSSize = m_fs[index]->FsSize;
            net_msg.UpdateProgress = FALSE;
            net_msg.xdt_trans(XDTOP_PUT, buffer, size);

            packet.SetData(buffer, size);
            if (! packet.Send(&socket))
            {
                succ = FALSE;
                continue;
            }
        }

        /*while ( 1 )
        {
            /* every 5 minutes send a beat to resource manager server * /
            sig_code = TimeWaitForRefReadySignal(300000);
            if (sig_code != 0)
                break;
            SendLoadUpdateToRM(11, m_used_time);
        }
        SendLoadUpdateToRM(11, m_used_time);*/

        /* wait 2 hours to refresh one file system */
        sig_code = TimeWaitForRefReadySignal(7200000);

        if (sig_code == 1)      /* ready, refreshing finished */
        {
            not_finish_num -= ref_num;
            continue;
        }
        else if (sig_code == 2) /* client error */
        {
            succ = FALSE;
            continue;
        }
        else    /* sig_code==3,4 (pause or stop) */
        {
            for (i=0; i<ref_num; i++)
            {
                if (! socket.ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT))
                {
                    printf("%s: Cannot connect to load client %s.\n", Safe_ctime(ctime_buf), m_clients[i]);
                    fflush(stdout);
                    succ = FALSE;
                    continue;
                }

                /* message type */
                net_msgtype.type = NET_MSGTYPE_PUMPER_STOP_REFFS;
                net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
                packet.SetData(buffer, size);
                if (! packet.Send(&socket))
                {
                    succ = FALSE;
                    continue;
                }

                net_loadid.id = m_load_id;
                net_loadid.xdt_trans(XDTOP_PUT, buffer, size);
                packet.SetData(buffer, size);
                if (! packet.Send(&socket))
                {
                    succ = FALSE;
                    continue;
                }
            }
            return (sig_code==3) ? 2 : 3;
        }
    }

    return succ ? 1 : 0;
}

/*********************** entry-point of this thread ***********************/

void CClientPrimMgrThread::run()
{
    char ctime_buf[32];

    run_once();

    printf("%s: Send quit message to all load clients\n", Safe_ctime(ctime_buf));
    fflush(stdout);
    SendQuitMessageClients();

    printf("%s: Prim client exit\n", Safe_ctime(ctime_buf));
    fflush(stdout);

    exit(1);
}

void CClientPrimMgrThread::run_once()
{
    fprintf(stderr, "In CClientPrimMgrThread::run()\n");

    int i, j;
    char error_msg[1024];
    char ctime_buf[32];
    int ref_fs_num, ref_fs_index[MAX_FSNUM];

    m_status = 0;   /* Initializing */

    /* read the load information from config file */
    printf("%s: Read config file (%s).\n", Safe_ctime(ctime_buf), m_ConfigFile);
    fflush(stdout);
    if (! ReadLoadInfo(error_msg))
    {
        printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
        fflush(stdout);
        return;
    }

    /* read the advanced setting from config file */
    CNet_LoadAdvance net_loadadv;
    char *adv_buffer;
    int adv_size;
    ReadAdvSetting(&net_loadadv, error_msg);
    net_loadadv.xdt_trans(XDTOP_PUT, adv_buffer, adv_size);

    /* then read the file system information from file sys info file */
    BOOL succ[MAX_FSNUM];
    printf("%s: Read file system information.\n", Safe_ctime(ctime_buf));
    fflush(stdout);
    if (! ReadFileSysInfo(succ))
    {
        ref_fs_num = 0;
        for (i=0; i<m_fsnum; i++)
        {
            if (! succ[i])
                ref_fs_index[ref_fs_num ++] = i;
        }

        printf("%s: Start to refresh %d file systems.\n", Safe_ctime(ctime_buf), ref_fs_num);
        fflush(stdout);
        m_status = 2;   /* Refreshing */
        int ref_result = RefeshFileSystems(ref_fs_num, ref_fs_index);
        m_status = 0;   /* Initializing */
        if (ref_result == 0)        /* Error */
        {
            strcpy(error_msg, "Automatic refresh file system error");
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }
        else if (ref_result == 1)   /* All finished */
        {
            /* SendLoadUpdateToRM(13, used_time);*/
            printf("%s: Finished to refresh file systems.\n", Safe_ctime(ctime_buf));
            fflush(stdout);
        }
        else                        /* Paused or stopped */
        {
            return;
        }
    }


    int lc_status[MAX_CLIENT_NUM];
    CSocketClient sock_to_clients[MAX_CLIENT_NUM];
    BOOL allMounted;
    int needRefresh;

    /* loop until all client see file system information is consistent with files on disk */
    BOOL loopTag = TRUE;
    while ( loopTag )
    {
        /* connect to all clients */
        printf("%s: Connect to %d load clients\n", Safe_ctime(ctime_buf), m_client_num);
        fflush(stdout);
        for (i=0; i<m_client_num; i++)
        {
            if (! sock_to_clients[i].ConnectToServer(m_client_ips[i], CLIENT_MAIN_PORT))
            {
                printf("%s: Fail to connect to load client %d: %s (%s).\n", Safe_ctime(ctime_buf), i+1, m_clients[i], m_client_ips[i]);
                fflush(stdout);

                /* Send a dummy message to all clients already connected to avoid blocking */
                CNet_MsgType net_msgtype;
                CPacket packet;
                char *buffer;
                int size;
                net_msgtype.type = NET_MSGTYPE_PUMPER_NONE;
                net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
                packet.SetData(buffer, size);
                for (j=0; j<i; j++)
                {
                    packet.Send(&sock_to_clients[j]);
                    packet.Receive(&sock_to_clients[j]);
                }

                return;
            }
            printf("%s: Success to connect to load client %d: %s (%s).\n", Safe_ctime(ctime_buf), i+1, m_clients[i], m_client_ips[i]);
            fflush(stdout);
        }

        /* now it's ready to start the load on each load client */
        printf("%s: Send start load command to all clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendStartLoadOnClients(sock_to_clients, adv_buffer, adv_size, error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* wait for the ready signal from all clients */
        printf("%s: Wait for ready signals from all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! WaitForLcReady(lc_status, 30000))
        {
            strcpy(error_msg, "Cannot receive ack from clients:");
            for (i=0; i<m_client_num; i++)
            {
                if (lc_status[i] < 0)
                {
                    strcat(error_msg, " ");
                    strcat(error_msg, m_clients[i]);
                }
            }
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* and to send file system base infomation to each load client */
        printf("%s: Send file system basic information to all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendFSBaseInfoToClients(sock_to_clients, error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* wait for the ready signal from all clients */
        /* clients need mount file system, then return ack, so here we wait more time (90 seconds) */
        printf("%s: Wait for ready signals from all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! WaitForLcReady(lc_status, 90000))
        {
            strcpy(error_msg, "Cannot receive ack from clients:");
            for (i=0; i<m_client_num; i++)
            {
                if (lc_status[i] < 0)
                {
                    strcat(error_msg, " ");
                    strcat(error_msg, m_clients[i]);
                }
            }
            strcat(error_msg, ". Timeout to mount file system.");
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* Check the load client's status (lc_status) to see file system is mounted correctly */
        allMounted = TRUE;
        strcpy(error_msg, "Some clients (");
        for (i=0; i<m_client_num; i++)
        {
            if (lc_status[i] != 0)
            {
                strcat(error_msg, " ");
                strcat(error_msg, m_clients[i]);
                allMounted = FALSE;
            }
        }
        strcat(error_msg, " ) can't mount file system");
        if (! allMounted)
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }
        printf("%s: Success to mount all file systems on all clients\n", Safe_ctime(ctime_buf));
        fflush(stdout);

        /* send go ahead signal to all load clients */
        printf("%s: Send go ahead signals to all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendGoAheadToClients(sock_to_clients, error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* shuffle the file systems */
        printf("%s: Shuffle all file systems.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        ShuffleAllFileSystems();

        /* send file system information to all clients */
        printf("%s: Send all file systems' information to all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendFileSysInfoToClients(sock_to_clients, error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* wait for the ready signal from all clients */
        /* clients need check file system, then return ack, so here we wait more time (60 seconds) */
        printf("%s: Wait for ready signals from all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! WaitForLcReady(lc_status, 60000))
        {
            strcpy(error_msg, "Cannot receive ack from clients:");
            for (i=0; i<m_client_num; i++)
            {
                if (lc_status[i] < 0)
                {
                    strcat(error_msg, " ");
                    strcat(error_msg, m_clients[i]);
                }
            }
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* Check the load client's status (lc_status) to see file system was refreshed correctly */
        needRefresh = 0;
        for (i=0; i<m_client_num; i++)
            needRefresh |= lc_status[i];
        ref_fs_num = 0;
        strcpy(error_msg, "Some file systems (");
        for (i=0; i<m_fsnum; i++)
        {
            if ((needRefresh & (1 << i)) != 0)
            {
                ref_fs_index[ref_fs_num ++] = i;
                strcat(error_msg, " ");
                strcat(error_msg, m_fs[i]->FsName);
            }
        }
        strcat(error_msg, " ) need to be refreshed");
        if (needRefresh == 0)
            loopTag = FALSE;
        else    /* some file system need to be refreshed */
        {
            /* print to log */
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);

            /* first close the connections to all client, let them timeout and quit */
            for (i=0; i<m_client_num; i++)
                sock_to_clients[i].ClosePort();

            /* then do refresh work */
            printf("%s: Start to refresh %d file systems.\n", Safe_ctime(ctime_buf), ref_fs_num);
            fflush(stdout);
            m_status = 2;   /* Refreshing */
            int ref_result = RefeshFileSystems(ref_fs_num, ref_fs_index);
            m_status = 0;   /* Initializing */
            if (ref_result == 0)        /* Error */
            {
                strcpy(error_msg, "Automatic refresh file system error");
                printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
                fflush(stdout);
                return;
            }
            else if (ref_result == 1)   /* All finished */
            {
                /* SendLoadUpdateToRM(13, used_time);*/
                printf("%s: Finished to refresh file systems.\n", Safe_ctime(ctime_buf));
                fflush(stdout);
            }
            else                        /* Paused or stopped */
            {
                return;
            }
        }
    }

    /* send go ahead signal to all load clients, they are running now */
    printf("%s: Send go ahead signals to all load clients.\n", Safe_ctime(ctime_buf));
    fflush(stdout);
    if (! SendGoAheadToClients(sock_to_clients, error_msg))
    {
        printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
        fflush(stdout);
        return;
    }

    /* send good news to RM with m_socket */
/*    if (! SendLoadRunResponseToRM(m_socket, TRUE, NULL))
        return;
*/
    printf("%s: Load is running on all load clients.\n", Safe_ctime(ctime_buf));
    fflush(stdout);

    /* close the connection to all client */
    for (i=0; i<m_client_num; i++)
        sock_to_clients[i].ClosePort();


    /*****************************************************/
    /*  here it's ready to go into run loop */
    /*****************************************************/

    int used_time = m_used_time;
    int sleep_time;
    int status;
    int sig_code;
    CSysTime t1(0, 0), t2(0, 0);

    m_status = 1;   /* Running */

    while (used_time < m_run_time)
    {
        if (used_time + m_shuffle_time <= m_run_time)
            sleep_time = m_shuffle_time;
        else
            sleep_time = m_run_time - used_time;

        //_SLEEP(sleep_time);

/*        total_sleep_time = 0;
        while (total_sleep_time < one_loop_sleep_time)
        {
            if (total_sleep_time + beat_time <= one_loop_sleep_time)
                sleep_time = beat_time;
            else
                sleep_time = one_loop_sleep_time - total_sleep_time;

            t1.GetCurTime();
            sig_code = TimeWaitForSignal(sleep_time * 1000);
            t2.GetCurTime();

            SendLoadUpdateToRM(10, m_used_time);

            if (sig_code == 0)  // timeout
                total_sleep_time += sleep_time;
            else                // woken up by signal (stop command is coming)
            {
                t2 -= t1;
                total_sleep_time += t2.m_sec;
                break;
            }
        }

        used_time += total_sleep_time;*/

        t1.GetCurTime();
        sig_code = TimeWaitForSignal(sleep_time * 1000);
        t2.GetCurTime();

        if (sig_code == 0)  /* timeout */
            used_time += sleep_time;
        else                /* woken up by signal (stop command is coming) */
        {
            t2 -= t1;
            used_time += t2.m_sec;
        }

        ResetFileSysTotalNum();
        ResetLcReady();

        if (sig_code > 0)  // pause or stop signal coming
            status = 2;
        else
            status = (used_time < m_run_time) ? 1 : 2;

        if (status == 1)
            printf("%s: Send break message to all load clients.\n", Safe_ctime(ctime_buf));
        else
            printf("%s: Send exit message to all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendBreakMessageToClients(status, error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* Wait for receiving updated file system information from all clients */
        /* here we need give load client time to wait for all ops finishing and organize data to return */
        printf("%s: Wait for ready signal from all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! WaitForLcReady(lc_status, 120000))
        {
            strcpy(error_msg, "Prim client is timeout, can't receiving all load clients stopped message.");
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        if (sig_code > 0 || used_time >= m_run_time)  // pause or stop signal coming, or time to finish
            break;

        /* Check and refresh file system */
        printf("%s: Check all file systems for refreshing.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        ref_fs_num = 0;
        for (i=0; i<m_fsnum; i++)
        {
            printf("%s: Check file systems %d (%s).\n", Safe_ctime(ctime_buf), i+1, m_fs[i]->FsName);
            fflush(stdout);
            if (CheckForRefresh(i))
                ref_fs_index[ref_fs_num ++] = i;
//DEBUG
//else
//ref_fs_index[ref_fs_num ++] = i;
        }
        if (ref_fs_num > 0)
        {
            printf("%s: Need to refresh %d file systems.\n", Safe_ctime(ctime_buf), ref_fs_num);
            fflush(stdout);
            m_used_time = used_time;
            m_status = 2;   /* Refreshing */
            int ref_result = RefeshFileSystems(ref_fs_num, ref_fs_index);
            m_status = 1;   /* Running */
            if (ref_result == 0)        /* Error */
            {
                strcpy(error_msg, "Automatic refresh file system error");
                printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
                fflush(stdout);
                return;
            }
            else if (ref_result == 1)   /* All finished */
            {
                /* SendLoadUpdateToRM(13, used_time);*/
                printf("%s: Finished to refresh file systems.\n", Safe_ctime(ctime_buf));
                fflush(stdout);
            }
            else                        /* Paused or stopped */
            {
                sig_code = (ref_result == 2) ? 1 : 2;
                break;
            }
        }

        /* shuffle the file systems again */
        printf("%s: Shuffle all file systems.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        ShuffleAllFileSystems();

        /* send updated file system information to all clients */
        printf("%s: Send all file systems' information to all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendFileSysInfoToClients(error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* wait for the ready signal from all clients */
        /* clients need check file system, then return ack, so here we wait more time (60 seconds) */
        printf("%s: Wait for ready signals from all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! WaitForLcReady(lc_status, 60000))
        {
            /* send back to RM with m_socket */
            strcpy(error_msg, "Cannot receive ack from clients:");
            for (i=0; i<m_client_num; i++)
            {
                if (lc_status[i] < 0)
                {
                    strcat(error_msg, " ");
                    strcat(error_msg, m_clients[i]);
                }
            }
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* Check the load client's status (lc_status) to see file system was refreshed correctly */
        needRefresh = 0;
        for (i=0; i<m_client_num; i++)
            needRefresh |= lc_status[i];
        strcpy(error_msg, "Some file systems (");
        for (i=0; i<m_fsnum; i++)
        {
            if ((needRefresh & (1 << i)) != 0)
            {
                strcat(error_msg, " ");
                strcat(error_msg, m_fs[i]->FsName);
            }
        }
        strcat(error_msg, " ) need to be refreshed");
        if (needRefresh != 0)
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        /* send go ahead signal to all load clients, they will go to next run */
        printf("%s: Send go ahead signals to all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        if (! SendGoAheadToClients(error_msg))
        {
            printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
            fflush(stdout);
            return;
        }

        printf("%s: Load is running on all load clients.\n", Safe_ctime(ctime_buf));
        fflush(stdout);
    
    }

    if (sig_code == 0)
        printf("%s: Load finished. Write new file system information.\n", Safe_ctime(ctime_buf));
    else if (sig_code == 1)
        printf("%s: Load paused. Write new file system information.\n", Safe_ctime(ctime_buf));
    else
        printf("%s: Load stoped. Write new file system information.\n", Safe_ctime(ctime_buf));
    fflush(stdout);

    if (! WriteFileSysInfo(error_msg))
    {
        printf("%s: %s.\n", Safe_ctime(ctime_buf), error_msg);
        fflush(stdout);
    }
}

/* for meta session */
void CClientPrimMgrThread::SetConfigFile(char *config_file)
{
    strcpy(m_ConfigFile, config_file);
}

BOOL CClientPrimMgrThread::ReadLoadInfo(char *error_msg)
{
    if (! m_config.ReadFromFile(m_ConfigFile, error_msg))
        return FALSE;

    m_load_id = 1;
    m_client_num = m_config.m_client_num;
    for (int i=0; i<m_client_num; i++)
    {
        strcpy(m_clients[i], m_config.m_clients[i]);
        strcpy(m_client_ips[i], m_config.m_client_ips[i]);
        m_load[i] = m_config.m_load[i];
    }
    m_stress_server = m_config.m_stress_server;
    m_run_time = m_config.m_run_time;
    m_used_time = 0;
    strcpy(m_rmname, "rm-host");
    strcpy(m_rmip, "1.1.1.1");
    strcpy(m_myname, m_config.m_primname);
    strcpy(m_myip, m_config.m_primip);

    return TRUE;
}

BOOL CClientPrimMgrThread::ReadAdvSetting(CNet_LoadAdvance *net_loadadv, char *error_msg)
{
    net_loadadv->stress_num = m_config.m_client_num;
    net_loadadv->loadlist_num = 0;
    net_loadadv->loadlist = NULL;
    net_loadadv->thread_num = m_config.m_thread_num;
    net_loadadv->opmix_num = (m_config.m_ostype == OSTYPE_UNIX) ? 17 : 13;
    net_loadadv->opmix = m_config.m_opmix;
    net_loadadv->mount_noac = TRUE;

    return TRUE;
}

BOOL CClientPrimMgrThread::ReadFileSysInfo(BOOL succ[])
{
    char run_dir[MAX_PATHLEN], *p;
    strcpy(run_dir, m_ConfigFile);
    p = strrchr(run_dir, _CHR_PATHSEP);
    if (p == NULL)
        run_dir[0] = '\0';
    else
        *p = '\0';

    m_fsnum = m_config.m_fs_num;

    int i, j;
    for (i=0; i<m_fsnum; i++)
        m_fs[i] = new SPrimFileSysInfo;

    CNet_FsInfo net_fsinfo[MAX_FSNUM];
    for (i=0; i<m_fsnum; i++)
        succ[i] = CFSInfoFile::ReadFileSysInfo(run_dir, i, &net_fsinfo[i]);

    /* receive file system information one by one */
    int subdirnum, dirmodenum, symlinkmodenum;
    BOOL allSucc = TRUE;
    for (i=0; i<m_fsnum; i++)
    {
        strcpy(m_fs[i]->ServerName, m_config.m_ServerName);
        strcpy(m_fs[i]->FsName, m_config.m_FsName[i]);
        strcpy(m_fs[i]->ServerNetIF, m_config.m_ServerNetIF[i]);
        strcpy(m_fs[i]->ServerPath, m_config.m_ServerPath[i]);
        strcpy(m_fs[i]->ShareName, m_config.m_ShareName[i]);
        m_fs[i]->FsSize = m_config.m_FsSize[i];

        if (! succ[i])
        {
            m_fs[i]->TotalSize = 0;
            m_fs[i]->TotalFNodeNum = 0;
            m_fs[i]->SubDirNum = 0;
            m_fs[i]->TotalIONum = 0;
            m_fs[i]->TotalFileNum = 0;
            m_fs[i]->TotalDirNum = 0;
            m_fs[i]->TotalSymlinkNum = 0;

            allSucc = FALSE;
            continue;
        }

        m_fs[i]->TotalSize = net_fsinfo[i].TotalSize;
        m_fs[i]->TotalFNodeNum = net_fsinfo[i].TotalFNodeNum;
        m_fs[i]->SubDirNum = net_fsinfo[i].SubDirNum;
        m_fs[i]->TotalIONum = net_fsinfo[i].TotalIONum;
        m_fs[i]->TotalFileNum = net_fsinfo[i].TotalFileNum;
        m_fs[i]->TotalDirNum = net_fsinfo[i].TotalDirNum;
        m_fs[i]->TotalSymlinkNum = net_fsinfo[i].TotalSymlinkNum;

        subdirnum = net_fsinfo[i].SubDirNum;

        m_fs[i]->IONum = new int [subdirnum];
        memcpy(m_fs[i]->IONum, net_fsinfo[i].IONum, subdirnum * sizeof(int));

        m_fs[i]->IOSize = new int * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            m_fs[i]->IOSize[j] = new int [m_fs[i]->IONum[j]];
            memcpy(m_fs[i]->IOSize[j], net_fsinfo[i].IOSize[j], m_fs[i]->IONum[j]*sizeof(int));
        }

        m_fs[i]->FileNum = new int [subdirnum];
        memcpy(m_fs[i]->FileNum, net_fsinfo[i].FileNum, subdirnum * sizeof(int));

        m_fs[i]->FileMode = new ushort * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            m_fs[i]->FileMode[j] = new ushort [m_fs[i]->FileNum[j]];
            memcpy(m_fs[i]->FileMode[j], net_fsinfo[i].FileMode[j], m_fs[i]->FileNum[j] * sizeof(ushort));
        }

        m_fs[i]->DirNum = new int [subdirnum];
        memcpy(m_fs[i]->DirNum, net_fsinfo[i].DirNum, subdirnum * sizeof(int));

        m_fs[i]->DirMode = new uchar * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            dirmodenum = SIZE_BIT_TO_BYTE(m_fs[i]->DirNum[j]);
            m_fs[i]->DirMode[j] = new uchar [dirmodenum];
            memcpy(m_fs[i]->DirMode[j], net_fsinfo[i].DirMode[j], dirmodenum * sizeof(uchar));
        }

        m_fs[i]->SymlinkNum = new int [subdirnum];
        memcpy(m_fs[i]->SymlinkNum, net_fsinfo[i].SymlinkNum, subdirnum * sizeof(int));

        m_fs[i]->SymlinkMode = new uchar * [subdirnum];
        for (j=0; j<subdirnum; j++)
        {
            symlinkmodenum = SIZE_BIT_TO_BYTE(m_fs[i]->SymlinkNum[j]);
            m_fs[i]->SymlinkMode[j] = new uchar [symlinkmodenum];
            memcpy(m_fs[i]->SymlinkMode[j], net_fsinfo[i].SymlinkMode[j], symlinkmodenum * sizeof(uchar));
        }

        m_fs[i]->OptDir = new int [m_fs[i]->SubDirNum];

        /* delete all spaces used by struct net_fsinfo[i] (new in CFSInfoFile::ReadFileSysInfo() ) */
        for (j=0; j<subdirnum; j++)
            delete [] net_fsinfo[i].IOSize[j];
        delete [] net_fsinfo[i].IOSize;
        delete [] net_fsinfo[i].IONum;

        for (j=0; j<subdirnum; j++)
            delete [] net_fsinfo[i].FileMode[j];
        delete [] net_fsinfo[i].FileMode;
        delete [] net_fsinfo[i].FileNum;

        for (j=0; j<subdirnum; j++)
            delete [] net_fsinfo[i].DirMode[j];
        delete [] net_fsinfo[i].DirMode;
        delete [] net_fsinfo[i].DirNum;

        for (j=0; j<subdirnum; j++)
            delete [] net_fsinfo[i].SymlinkMode[j];
        delete [] net_fsinfo[i].SymlinkMode;
        delete [] net_fsinfo[i].SymlinkNum;
    }

    return allSucc;
}

BOOL CClientPrimMgrThread::WriteFileSysInfo(char *error_msg)
{
    char run_dir[MAX_PATHLEN], *p;
    strcpy(run_dir, m_ConfigFile);
    p = strrchr(run_dir, _CHR_PATHSEP);
    if (p == NULL)
        run_dir[0] = '\0';
    else
        *p = '\0';

    CNet_FsInfo net_fsinfo;
    for (int i=0; i<m_fsnum; i++)
    {
        net_fsinfo.TotalSize        = m_fs[i]->TotalSize;
        net_fsinfo.TotalFNodeNum    = m_fs[i]->TotalFNodeNum;
        net_fsinfo.SubDirNum        = m_fs[i]->SubDirNum;
        net_fsinfo.TotalIONum       = m_fs[i]->TotalIONum;
        net_fsinfo.TotalFileNum     = m_fs[i]->TotalFileNum;
        net_fsinfo.TotalDirNum      = m_fs[i]->TotalDirNum;
        net_fsinfo.TotalSymlinkNum  = m_fs[i]->TotalSymlinkNum;
        net_fsinfo.IONum            = m_fs[i]->IONum;
        net_fsinfo.IOSize           = m_fs[i]->IOSize;
        net_fsinfo.FileNum          = m_fs[i]->FileNum;
        net_fsinfo.FileMode         = m_fs[i]->FileMode;
        net_fsinfo.DirNum           = m_fs[i]->DirNum;
        net_fsinfo.DirMode          = m_fs[i]->DirMode;
        net_fsinfo.SymlinkNum       = m_fs[i]->SymlinkNum;
        net_fsinfo.SymlinkMode      = m_fs[i]->SymlinkMode;

        if (! CFSInfoFile::WriteFileSysInfo(run_dir, i, &net_fsinfo))
        {
            sprintf(error_msg, "Cannot write file system information (%d) to file", i+1);
            return FALSE;
        }
    }

    return TRUE;
}
