
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ConfigFile.h
**
** DESCRIPTION: This is header file to define the class:
**              CConfigFile
**
**              This class mainly manager config file.
**              
**
** AUTHOR: Yubing Wang       Mar-01-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
**  CLASS LIST:
**      CConfigFile
**
**
****************************************************************************/

#ifndef _CONFIGFILE_H_
#define _CONFIGFILE_H_

#include "Common.h"

#define MAX_CLIENT_NUM      64          /* max number of client for one load */


#define MAX_ARGNANE_LEN       32

typedef struct _ARG_TYPE
{
    char name[MAX_ARGNANE_LEN];
    char *value;
} ARG_TYPE;

#define CONFIG_ARG_PRIM_NAME        0
#define CONFIG_ARG_PRIM_IP          1
#define CONFIG_ARG_LOAD_NAME        2
#define CONFIG_ARG_LOAD_IP          3
#define CONFIG_ARG_STRESS_SERVER    4
#define CONFIG_ARG_LOAD_LEVEL       5
#define CONFIG_ARG_RUN_TIME         6
#define CONFIG_ARG_PROTOCOL         7
#define CONFIG_ARG_SERVER_NAME      8
#define CONFIG_ARG_FS_MNT_POINT     9
#define CONFIG_ARG_FS_SIZE          10
#define CONFIG_ARG_THREAD_NUM       11
#define CONFIG_ARG_OPMIX_FILE       12
#define CONFIG_ARG_NUMBER           CONFIG_ARG_OPMIX_FILE+1

#define OPMIX_ARG_Read              0
#define OPMIX_ARG_OverWrite         1
#define OPMIX_ARG_Append            2
#define OPMIX_ARG_Truncate          3
#define OPMIX_ARG_Stat              4
#define OPMIX_ARG_Chmod             5
#define OPMIX_ARG_CreatNULL         6
#define OPMIX_ARG_Mkdir             7
#define OPMIX_ARG_Rmdir             8
#define OPMIX_ARG_ReadDir           9
#define OPMIX_ARG_Symlink           10
#define OPMIX_ARG_Readlink          11
#define OPMIX_ARG_Link              12
#define OPMIX_ARG_Unlink            13
#define OPMIX_ARG_Rename            14
#define OPMIX_ARG_Statvfs           15
#define OPMIX_ARG_Memmap            16
#define OPMIX_ARG_GetAttr           17
#define OPMIX_ARG_SetAttr           18
#define OPMIX_ARG_Delete            19
#define OPMIX_ARG_GetDiskSpace      20
#define OPMIX_ARG_NUMBER            OPMIX_ARG_GetDiskSpace+1

class CConfigFile
{
public:
    char m_primname[MAX_NAMELEN];
    char m_primip[MAX_IPADDRLEN];
    int m_client_num;
    char m_clients[MAX_CLIENT_NUM][MAX_NAMELEN];
    char m_client_ips[MAX_CLIENT_NUM][MAX_IPADDRLEN];
    int m_load[MAX_CLIENT_NUM];
    BOOL m_stress_server;
    int m_run_time;
    int m_ostype;

    int m_thread_num;
    int m_opmix[32];

    char m_ServerName[MAX_NAMELEN*2];
    int m_fs_num;
    char m_FsName[MAX_FSNUM][MAX_NAMELEN];           /* file system name */
    char m_ServerNetIF[MAX_FSNUM][MAX_NAMELEN];      /* server network interface name or ip */
    char m_ServerPath[MAX_FSNUM][MAX_PATHLEN];       /* server: the path that file system mount on */
    char m_ShareName[MAX_FSNUM][MAX_NAMELEN];        /* server: the shared name of file system for CIFS protocol */
    int m_FsSize[MAX_FSNUM];
    

private:
    ARG_TYPE config_args[CONFIG_ARG_NUMBER];
    ARG_TYPE opmix_args[OPMIX_ARG_NUMBER];

public:
    CConfigFile();
    ~CConfigFile();

    /* functions to manage arguments */
    void init_arguments();
    void free_arguments();
    BOOL read_arguments(char *file_name, int args_num, ARG_TYPE args[], char *error_msg);
    // rewrite fgets function to read the unix format text file in dos enviroment
    char *fgets_unix(char *string, int n, FILE *stream);
    BOOL read_one(char *line, int args_num, ARG_TYPE args[]);
    void resign_value(int arg_index, ARG_TYPE args[], char *value);

    /* fucntions to manage prim client config file */
    BOOL ReadFromFile(char *file_name, char *error_msg);
    BOOL GetConfigPara(char *error_msg);
    void SetDefaultOpmix();
    BOOL GetOpmix(char *error_msg);
};

#endif
