
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysBase.h
**
** DESCRIPTION: This is header file to define the class:
**              CPumper
**
**              This class implement the main procedure of the pumper program.
**              
**
** AUTHOR: Yubing Wang       Mar-30-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
**  CLASS LIST:
**      CPumper
**
**
****************************************************************************/

#ifndef _PUMPER_H_
#define _PUMPER_H_

#include "Common.h"

class CBaseThread;

class CThreadList
{
private:
    int m_Type;     /* 1:CClientFileSysThread 2:CClientPrimMgrThread, 3:CClientLoadMgrThread */
    int m_ThreadNum;
    int m_MaxThreadNum;
    CBaseThread **m_Threads;

#ifndef WIN32
    pthread_mutex_t m_lock;
#else
    HANDLE m_lock;
#endif

/* constructor and destructor */
public:
    CThreadList(int type);
    ~CThreadList();

/* operations */
public:
    void AddThread(CBaseThread *thread);
    void DeleteThread(CBaseThread *thread);
    CBaseThread *FindThreadByLoadID(int id);
};


#ifndef WIN32

#define MAX_CLIENT_FSNUM        32

/* this class is only for unix version */
class CLocalPathManager
{
private:
    BOOL m_LocalPathFree[MAX_CLIENT_FSNUM];
    pthread_mutex_t m_lock;

/* constructor and destructor */
public:
    CLocalPathManager();
    ~CLocalPathManager();

/* operations */
public:
    void ApplyLocalPath(char *localpath);
    void FreeLocalPath(char *localpath);
};

#endif

class CPumper
{
private:
    CThreadList m_FileSysThreadList;
    CThreadList m_PrimThreadList;
    CThreadList m_LoadThreadList;

public:
    int opnum;

/* constructor and destructor */
public:
    CPumper();
    ~CPumper();

/* operations */
public:
    int AppMain(int option, char *config_file, char *prim_name);
    void WaitForRequest();
    void StopLoad(char *prim_name);

    void AddThread(int type, CBaseThread *thread);
    void DeleteThread(int type, CBaseThread *thread);
    CBaseThread *FindThreadByLoadID(int type, int id);
};

#endif
